# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""GUI for edition of templates for inputs and outputs files To be used by
:class:`.DiscFromExe`

Run this file with no argument to open the GUI
"""
from __future__ import division, unicode_literals

import logging
import sys
from os.path import dirname, exists, join

LOGGER = logging.getLogger(__name__)

try:
    from PySide2.QtCore import QRegExp
    from PySide2.QtGui import QColor, QIcon, QTextCursor
    from PySide2.QtWidgets import (
        QAction,
        QApplication,
        QFileDialog,
        QInputDialog,
        QMainWindow,
        QTextEdit,
    )
except ImportError:
    # Both libraries are fully interchangeable so we are not contaminated by
    # the GPL license here
    LOGGER.warning("PySide2 cannot be imported.")
    from PyQt5.QtCore import QRegExp
    from PyQt5.QtGui import QColor, QIcon, QTextCursor
    from PyQt5.QtWidgets import (
        QAction,
        QApplication,
        QFileDialog,
        QInputDialog,
        QMainWindow,
        QTextEdit,
    )

    LOGGER.warning(
        "Your Python environment uses PyQt5"
        ": it is distributed under the GNU GPL v3.0 license "
        "unless you have acquired a commercial license for it."
    )


class QtTemplateEditor(QMainWindow):
    """GUI template generator.

    GUI to generate templates for input and output files edition
    Input text file data is replaced by a mark that specifies where to read the
    data.
    idem for outputs.
    Works with any text file format.

    To open the GUI, run this python file

    Example, for an input json file :
    {
    "a": 1.01515112125,
    "b": 2.00151511213,
    "c": 3.00151511213
    }

    Generates a template :
    {
    "a": GEMSEO_INPUT{a::1.0},
    "b": GEMSEO_INPUT{b::2.0},
    "c": GEMSEO_INPUT{c::3.0}
    }

    Same for outputs.
    """

    def __init__(self, in_sep="GEMSEO_INPUT", out_sep="GEMSEO_OUTPUT"):
        """Constructor :

        :param in_sep: separator name for the input tag, default GEMSEO_INPUT
        :param out_sep: separator name for the output tag, default GEMSEO_OUTPUT
        """

        self.in_sep = in_sep
        self.out_sep = out_sep

        QMainWindow.__init__(self, None)
        self.q_text_e = QTextEdit(self)
        self._setup_toolbars()

        self.q_text_e.setTabStopWidth(12)
        self.setCentralWidget(self.q_text_e)

        self.setGeometry(100, 100, 600, 800)

    def _setup_toolbars(self):
        """Setup the toolbars, the icons and shortcuts."""
        self.toolbar = self.addToolBar("Actions")

        self.add_action("Open", "Open existing document", "Ctrl+Shift+O", self.open_doc)
        self.add_action("Save", "Save document", "Ctrl+S", self.save_doc)
        self.toolbar.addSeparator()

        self.add_action("Input", "Make input", "Ctrl+I", self.make_input)
        self.add_action("Output", "Make output", "Ctrl+O", self.make_output)

        self.toolbar.addSeparator()

        self.add_action("Copy", "Copy text to clipboard", "Ctrl+C", self.q_text_e.copy)
        self.add_action(
            "Paste", "Paste text from clipboard", "Ctrl+V", self.q_text_e.paste
        )
        self.add_action("Cut", "Cut text from clipboard", "Ctrl+X", self.q_text_e.cut)

        self.toolbar.addSeparator()

        self.add_action("Undo", "Undo last action", "Ctrl+Z", self.q_text_e.undo)
        self.add_action("Redo", "Redo last action", "Ctrl+Shift+Z", self.q_text_e.redo)

        self.toolbar.addSeparator()

        self.addToolBarBreak()

    def __get_open_filename(self, name):
        """Open a dialog to select a file to open."""
        filename, _ = QFileDialog.getOpenFileName(self, name)
        is_not_empty = bool(filename)
        if is_not_empty and filename is not None:
            return filename, True
        return "", False

    def __get_save_filename(self, name):
        """Open a dialog to select a file to save."""
        filename, _ = QFileDialog.getSaveFileName(self, name)
        is_not_empty = bool(filename)
        if is_not_empty and filename is not None:
            return filename, True
        return "", False

    def add_action(self, name, status_tip, shortcut, connect):
        """Add an action with a button and icon.

        :param name: name of the action
        :param status_tip: tip for the user to browse with the mouse
        :param shortcut: keyboard shortcut (Ctrl+S) for instance
        :param connect: method to call at trigger
        """
        icon_path = join(dirname(__file__), "icons", name + ".png")
        if exists(icon_path):
            icon = QIcon(icon_path)
            action = QAction(icon, name, self)
        else:
            action = QAction(name, self)
        action.setStatusTip(status_tip)
        action.setShortcut(shortcut)
        action.triggered.connect(connect)
        self.toolbar.addAction(action)
        return action

    def open_doc(self):
        """Open the document for edition of the template."""
        filename, is_ok = self.__get_open_filename("Open File")
        if is_ok:
            f_handler = open(filename, "r")
            filedata = f_handler.read()
            self.q_text_e.setText(filedata)
            f_handler.close()
            self.hightlight(self.in_sep, "green")
            self.hightlight(self.out_sep)

    def save_doc(self):
        """Save the template to a file."""
        filename, is_ok = self.__get_save_filename("Save File")
        if is_ok:
            f_handler = open(filename, "w")
            filedata = self.q_text_e.toPlainText()
            f_handler.write(filedata)
            f_handler.close()

    def make_input(self):
        """Make an input from the selected data."""
        name, is_ok = QInputDialog.getText(self, "Input name", "Enter the input name:")
        if is_ok:
            name = str(name)
            cursor = self.q_text_e.textCursor()

            selection = cursor.selection().toPlainText()

            tag = self.in_sep + "{" + name + "::" + selection + "}"
            cursor.insertText(tag)
            self.hightlight(self.in_sep, "green")

    def make_output(self):
        """Make an output from the selected data."""
        name, is_ok = QInputDialog.getText(
            self, "Output name", "Enter the output name:"
        )
        if is_ok:
            name = str(name)
            cursor = self.q_text_e.textCursor()
            selection = cursor.selection().toPlainText()
            tag = self.out_sep + "{" + name + "::" + selection + "}"
            cursor.insertText(tag)
            self.hightlight(self.out_sep)

    def hightlight(self, sep, color="red"):
        """Highight some text.

        :param sep: the regex that validates the text to highlight
        :param color: the color to be used
        """
        # Setup the desired format for matches
        color = QColor(color)
        color.setAlpha(100)
        # Setup the regex engine
        regex = QRegExp(sep)
        # Process the displayed document
        index = regex.indexIn(self.q_text_e.toPlainText(), 0)
        cursor = self.q_text_e.textCursor()

        while index != -1:
            # Select the matched text and apply the desired format
            cursor.setPosition(index)
            cursor.movePosition(QTextCursor.EndOfWord, QTextCursor.KeepAnchor, n=1)
            charfmt = cursor.charFormat()
            charfmt.setBackground(color)
            cursor.setCharFormat(charfmt)

            # Move to the next match
            pos = index + regex.matchedLength()
            index = regex.indexIn(self.q_text_e.toPlainText(), pos)


def main():
    """Entry point."""
    app = QApplication(sys.argv)
    editor = QtTemplateEditor()
    editor.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Excel based discipline."""

from __future__ import division, unicode_literals

import atexit
import os
from os import getcwd

from numpy import array

try:
    cwd = getcwd()
    import xlwings
except ImportError:
    # error will be reported if the discipline is used
    os.chdir(cwd)
    xlwings = None

from gemseo.core.discipline import MDODiscipline


class XLSDiscipline(MDODiscipline):
    """Wraps an excel workbook into a discipline.

    .. warning::

        As this wrapper relies on the `xlswings library <https://www.xlwings.org>`__
        to handle macros and interprocess communication, it
        is only working under Windows and MacOS.
    """

    def __init__(self, xls_file_path, macro_name="execute"):
        """Initialize xls file path and macro.

        Inputs must be specified in the "Inputs" sheet, in the following
        format (A and B are the first two columns)

        +---+---+
        | A | B |
        +===+===+
        | a | 1 |
        +---+---+
        | b | 2 |
        +---+---+

        Where a is the name of the first input, and 1 is its default value
        b is the name of the second one, and 2 its default value.

        There must be no empty lines between the inputs


           The number of rows is arbitrary but they must be contiguous and
           start at line 1

           And same for the "Outputs" sheet (A and B are the first two columns)


        +---+---+
        | A | B |
        +===+===+
        | c | 3 |
        +---+---+

        Where c is the only output. There may be multiple.

           if the file is a .xlsm, a macro named "execute" *must* exist
           and will be called by the _run method before retrieving
           the outputs. The macro has no arguments, it takes its inputs in the
           Inputs sheet, and write the outputs to the "Outputs" sheet

           Alternatively, the user may provide a macro name to the constructor,
           or None if no macro shall be executed


        Parameters
        ----------
        xls_file_path : str
            path to the excel file
            if the file is a .xlsm, a macro named "execute" must exist
            and will be called by the _run method before retrieving
            the outputs
        macro_name : str
            name of the macro to be executed for a .xlsm file
            if None is provided, do not run a macro
        """
        if xlwings is None:
            raise ImportError("cannot import xlwings")

        self._book = None

        try:
            self._xls_app = xlwings.App(visible=False)
        # wide except because I cannot tell what is the exception raised by xlwings
        except:  # noqa: E722,B001
            raise RuntimeError("xlwings requires Microsoft Excel")

        super(XLSDiscipline, self).__init__()
        self._xls_file_path = xls_file_path
        self.macro_name = macro_name

        # Close the app when exiting
        atexit.register(self._xls_app.quit)

        self._book = xlwings.Book(xls_file_path)
        sh_names = [sheet.name for sheet in self._book.sheets]
        if "Inputs" not in sh_names:
            raise ValueError(
                "Workbook must contain a sheet named 'Inputs' "
                + "that define the inputs of the discipline"
            )
        if "Outputs" not in sh_names:
            raise ValueError(
                "Workbook must contain a sheet named 'Outputs' "
                + "that define the outputs of the discipline"
            )
        self.input_names = None
        self.output_names = None

        self._init_grammars()
        self._init_defaults()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def close(self):
        """Close the workbook."""
        if self._book is not None:
            self._book.close()
            self._book = None

    def __del__(self):
        # ensures that the gc automatically release the resources held by the book
        self.close()

    def __read_sheet_col(self, sheet_name, column=0):
        """Read a specific column of the sheet."""
        sht = self._book.sheets[sheet_name]
        i = 0
        value = sht[i, column].value
        values = []
        while value is not None:
            values.append(value)
            i += 1
            value = sht[i, column].value

        return values

    def _init_grammars(self):
        """Initialize grammars by parsing the Inputs and Outputs sheets."""
        self.input_names = self.__read_sheet_col("Inputs", 0)
        self.output_names = self.__read_sheet_col("Outputs", 0)
        self.input_grammar.initialize_from_data_names(self.input_names)
        self.output_grammar.initialize_from_data_names(self.output_names)

    def _init_defaults(self):
        """Initialize the default input values."""
        inputs = self.__read_sheet_col("Inputs", 1)
        if len(inputs) != len(self.input_names):
            msg = (
                "Inconsistent Inputs sheet, names (first columns) and"
                " values column (second) must be of the same length"
            )
            raise ValueError(msg)

        self.default_inputs = dict(zip(self.input_names, inputs))

    def __write_inputs(self, input_data):
        """Write the inputs values to the Inputs sheet."""
        sht = self._book.sheets["Inputs"]
        for i, key in enumerate(self.input_names):
            sht[i, 1].value = input_data[key][0]

    def _run(self):
        """Run the discipline.

        Eventually calls the execute macro
        """
        self.__write_inputs(self.local_data)
        if self._xls_file_path.endswith(".xlsm") and self.macro_name is not None:
            try:
                self._xls_app.api.Application.Run(self.macro_name)
            except Exception as err:
                macro_name = self.macro_name
                msg = "Failed to run '{}' macro: {}.".format(macro_name, err)
                raise RuntimeError(msg)

        out_vals = self.__read_sheet_col("Outputs", 1)
        if len(out_vals) != len(self.output_names):
            msg = (
                "Inconsistent Outputs sheet, names (first columns) and "
                " values column (second) must be of the same length."
            )
            raise ValueError(msg)

        outputs = {k: array([v]) for k, v in zip(self.output_names, out_vals)}
        self.store_local_data(**outputs)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Benoit Pauwels
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Factory for the trust updater
*****************************
"""
from __future__ import division, unicode_literals

import logging

from gemseo.algos.opt.core.trust_updater import PenaltyUpdater, RadiusUpdater

LOGGER = logging.getLogger(__name__)


class UpdaterFactory(object):
    """Creates the trust updater."""

    RADIUS = "radius"
    PENALTY = "penalty"
    TRUST_PARAMETERS = [RADIUS, PENALTY]

    def __init__(self):
        """Initializer."""
        self.__update_name_to_updater = {
            UpdaterFactory.RADIUS: RadiusUpdater,
            UpdaterFactory.PENALTY: PenaltyUpdater,
        }

    def create(self, name, thresholds, multipliers, bound):
        """Factory method to create a TrustUpdater subclass from an update name.

        :param name: update name
        :type name: string
        :param thresholds: thresholds for the decreases ratio
        :type thresholds: tuple
        :param multipliers: multipliers for the trust parameter
        :type multipliers: tuple
        :param bound: (lower or upper) bound for the trust parameter
        :returns: trust updater
        """
        if name in self.__update_name_to_updater:
            updater = self.__update_name_to_updater[name](
                thresholds, multipliers, bound
            )
        else:
            raise ValueError(
                "No update method named "
                + str(name)
                + " is available among update methods : "
                + str(list(self.__update_name_to_updater.keys()))
            )
        return updater



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Singletons implementation and variants
**************************************
"""
from __future__ import division, unicode_literals

from os.path import realpath

from gemseo.utils.py23_compat import string_types


class SingleInstancePerAttributeId(type):
    """A Singleton-like design pattern so that subclasses are only instantiated when the
    discipline instance passed as input of the constructor is different from already
    created instances.

    The test if the instances are equal is made with the id(obj1)==id(obj2) operator
    """

    instances = {}

    # Eclipse is not happy with "cls" as first
    # argument but this is an eclipse bug.
    # "function.MDOFunctionGenerator' should have self as first parameter"
    def __call__(cls, *args, **kwargs):
        # id = memory address of the object, which is unique
        if not args:
            raise ValueError(
                "SingleInstancePerAttribute subclasses "
                + "need at"
                + " least one attribute in the constructor."
            )
        inst_key = (id(cls), id(args[0]))
        inst = cls.instances.get(inst_key)
        if inst is None:
            inst = type.__call__(cls, *args, **kwargs)
            cls.instances[inst_key] = inst
        return inst


class SingleInstancePerAttributeEq(type):
    """A Singleton-like design pattern so that subclasses are only instantiated when the
    discipline instance passed as input of the constructor is different from already
    created instances.

    The test if the instances are equal is made with the obj1 == obj2 operator
    """

    instances = {}

    # Eclipse is not happy with "cls" as first
    # argument but this is an eclipse bug.
    # "function.MDOFunctionGenerator' should have self as first parameter"
    def __call__(cls, *args, **kwargs):
        if not args:
            args = [None]
        inst_key = (id(cls),) + tuple(args)
        inst = cls.instances.get(inst_key)
        if inst is None:
            inst = type.__call__(cls, *args, **kwargs)
            cls.instances[inst_key] = inst
        return inst


class SingleInstancePerFileAttribute(type):
    """A Singleton-like design pattern so that subclasses are only instantiated when the
    discipline instance passed as input of the constructor is different from already
    created instances.

    The test if the instances are equal is made with the obj1 == obj2 operator
    """

    instances = {}

    # Eclipse is not happy with "cls" as first
    # argument but this is an eclipse bug.
    def __call__(cls, *args, **kwargs):
        if not args:
            raise ValueError(
                "SingleInstancePerAttribute subclasses need at"
                + " least one attribute in the constructor."
            )
        fpath = args[0]

        if not isinstance(fpath, string_types):
            raise TypeError(
                "Argument 0 is not a string but of type :" + str(type(fpath))
            )
        fpath = realpath(fpath)
        inst_key = (id(cls), fpath)
        inst = cls.instances.get(inst_key)
        if inst is None:
            inst = type.__call__(cls, *args, **kwargs)
            cls.instances[inst_key] = inst
        return inst



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Damien Guenot - 26 avr. 2016
#       :author: Francois Gallard, refactoring
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""
Driver library
==============

A driver library aims to solve an :class:`.DriverLib`
using a particular algorithm from a particular family of numerical methods.
This algorithm will be in charge of evaluating the objective and constraints
functions at different points of the design space, using the
:meth:`.DriverLib.execute` method.
The most famous kinds of numerical methods to solve an optimization problem
are optimization algorithms and design of experiments (DOE). A DOE driver
browses the design space agnostically, i.e. without taking into
account the function evaluations. On the contrary, an optimization algorithm
uses this information to make the journey through design space
as relevant as possible in order to reach as soon as possible the optimum.
These families are implemented in :class:`.DOELibrary`
and :class:`.OptimizationLibrary`.
"""

from __future__ import division, unicode_literals

import inspect
import io
import logging
import string
from os.path import dirname, exists, join
from time import time

import tqdm
from numpy import ones_like, where, zeros_like

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.algos.opt_result import OptimizationResult
from gemseo.algos.stop_criteria import (
    DesvarIsNan,
    FtolReached,
    FunctionIsNan,
    MaxIterReachedException,
    MaxTimeReached,
    TerminationCriterion,
    XtolReached,
)
from gemseo.core.grammar import InvalidDataException
from gemseo.core.json_grammar import JSONGrammar
from gemseo.utils.source_parsing import get_options_doc

LOGGER = logging.getLogger(__name__)


class TqdmToLogger(io.StringIO):
    """Redirect tqdm output to the gemseo logger."""

    def write(self, buf):
        """Write buffer."""
        buf = buf.strip(string.whitespace)
        if buf:
            LOGGER.info(buf)


class ProgressBar(tqdm.tqdm):
    """Extend tqdm progress bar with better time units.

    Use hour, day or week for slower processes.
    """

    @classmethod
    def format_meter(cls, n, total, elapsed, **kwargs):
        if elapsed != 0.0:
            rate, unit = cls.__convert_rate(total, elapsed)
            kwargs["rate"] = rate
            kwargs["unit"] = unit
        meter = tqdm.tqdm.format_meter(n, total, elapsed, **kwargs)
        # remove the unit suffix that is hard coded in tqdm
        return meter.replace("/s,", ",").replace("/s]", "]")

    @staticmethod
    def __convert_rate(total, elapsed):
        rps = float(total) / elapsed
        if rps >= 0:
            rate = rps
            unit = "sec"

        rpm = rps * 60.0
        if rpm < 60.0:
            rate = rpm
            unit = "min"

        rph = rpm * 60.0
        if rph < 60.0:
            rate = rph
            unit = "hour"

        rpd = rph * 24.0
        if rpd < 24.0:
            rate = rpd
            unit = "day"

        return rate, " it/{}".format(unit)


class DriverLib(object):
    """Abstract class for DOE & optimization libraries interfaces.

    Lists available methods in the library for the proposed
    problem to be solved.

    To integrate an optimization package, inherit from this class
    and put your file in gemseo.algos.doe or gemseo.algo.opt packages.
    """

    USER_DEFINED_GRADIENT = OptimizationProblem.USER_GRAD
    COMPLEX_STEP_METHOD = OptimizationProblem.COMPLEX_STEP
    FINITE_DIFF_METHOD = OptimizationProblem.FINITE_DIFFERENCES

    DIFFERENTIATION_METHODS = [
        USER_DEFINED_GRADIENT,
        COMPLEX_STEP_METHOD,
        FINITE_DIFF_METHOD,
    ]

    LIB = "lib"
    INTERNAL_NAME = "internal_algo_name"
    OPTIONS_DIR = "options"
    OPTIONS_MAP = {}
    PROBLEM_TYPE = "problem_type"
    REQUIRE_GRAD = "require_grad"
    HANDLE_EQ_CONS = "handle_equality_constraints"
    HANDLE_INEQ_CONS = "handle_inequality_constraints"
    POSITIVE_CONSTRAINTS = "positive_constraints"
    INEQ_TOLERANCE = "ineq_tolerance"
    EQ_TOLERANCE = "eq_tolerance"
    MAX_TIME = "max_time"
    USE_DATABASE_OPTION = "use_database"
    NORMALIZE_DESIGN_SPACE_OPTION = "normalize_design_space"
    ROUND_INTS_OPTION = "round_ints"
    WEBSITE = "website"
    DESCRIPTION = "description"
    MAX_DS_SIZE_PRINT = 40

    def __init__(self):
        """Constructor."""
        # Library settings and check
        self.lib_dict = {}
        self.algo_name = None
        self.internal_algo_name = None
        self.problem = None
        self.opt_grammar = None
        self.__progress_bar = None
        self.__max_iter = 0
        self.__iter = 0
        self._start_time = None
        self._max_time = None

    def init_options_grammar(self, algo_name):
        """Initializes the options grammar.

        :param algo_name: name of the algorithm
        """
        # Try algo name convention which has the prioryty over
        # library options
        basename = algo_name.upper() + "_options.json"
        lib_dir = inspect.getfile(self.__class__)
        opt_dir = join(dirname(lib_dir), self.OPTIONS_DIR)
        algo_schema_file = join(opt_dir, basename)
        if exists(algo_schema_file):
            schema_file = algo_schema_file
        else:
            # Try to load library options convention by default
            basename = self.__class__.__name__.upper() + "_options.json"
            lib_schema_file = join(opt_dir, basename)

            if exists(lib_schema_file):
                schema_file = lib_schema_file
            else:
                msg = "Options grammar file " + algo_schema_file
                msg += " for algorithm: " + algo_name
                msg += " not found. And library options grammar file "
                msg += lib_schema_file + " not found either."
                raise ValueError(msg)

        descr_dict = get_options_doc(self.__class__._get_options)
        self.opt_grammar = JSONGrammar(
            algo_name, schema_file=schema_file, descriptions=descr_dict
        )

        return self.opt_grammar

    @property
    def algorithms(self):
        """Return the available algorithms."""
        return list(self.lib_dict.keys())

    def init_iter_observer(self, max_iter, message):
        """Initialize the iteration observer.

        It will handle the stopping criterion and the logging of the progress bar.

        :param max_iter: maximum number of calls
        :param message: message to display at the beginning
        """
        if max_iter < 1:
            raise ValueError("max_iter must be >=1, got {}".format(max_iter))
        self.__max_iter = max_iter
        self.__iter = len(self.problem.database)
        self.__progress_bar = ProgressBar(
            total=self.__max_iter,
            desc=message,
            ascii=False,
            file=TqdmToLogger(),
        )

    def __set_progress_bar_objective_value(self):
        value = self.problem.objective.last_eval
        if value is not None:
            # if maximization problem: take the opposite
            if not self.problem.minimize_objective:
                value = -value
            self.__progress_bar.set_postfix(refresh=False, obj=value)

    def new_iteration_callback(self):
        """Callback called at each new iteration, ie every time a design vector that is
        not already in the database is proposed by the optimizer.

        Iterates the progress bar, implements the stop criteria.
        """
        # First check if the max_iter is reached and update the progress bar
        self.__iter += 1
        if self.__iter > self.__max_iter:
            raise MaxIterReachedException()
        if self._max_time > 0:
            delta_t = time() - self._start_time
            if delta_t > self._max_time:
                raise MaxTimeReached()

        self.__set_progress_bar_objective_value()
        self.__progress_bar.update()

    def finalize_iter_observer(self):
        """Finalize the iteration observer."""
        self.__set_progress_bar_objective_value()
        self.__progress_bar.close()

    def _pre_run(self, problem, algo_name, **options):
        """To be overridden by subclasses Specific method to be executed just before
        _run method call.

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
        :param options: the options dict for the algorithm, see associated JSON file
        """
        self._start_time = time()
        self._max_time = options.get(self.MAX_TIME, 0.0)

    def _post_run(self, problem, algo_name, result, **options):  # pylint: disable=W0613
        """To be overridden by subclasses Specific method to be executed just after _run
        method call.

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
        :param result: result of the run such as an OptimizationResult
        :param options: the options dict for the algorithm, see associated JSON file
        """
        LOGGER.info("%s", result)
        problem.solution = result
        if result.x_opt is not None:
            problem.design_space.set_current_x(result)
        if problem.design_space.dimension <= self.MAX_DS_SIZE_PRINT:
            LOGGER.info("%s", problem.design_space)

    def driver_has_option(self, option_key):
        """Checks if the option key exists.

        :param option_key: the name of the option
        :return: True if the option is in the grammar
        """
        return self.opt_grammar.is_data_name_existing(option_key)

    def _process_options(self, **options):
        """After _get_options is called, the options are converted to algorithm specific
        options, and checked.

        :param options: driver options
        """
        for option_key in list(options.keys()):  # Copy keys on purpose
            # Remove extra options added in the _get_option method of the
            # driver
            if not self.driver_has_option(option_key):
                del options[option_key]
            elif option_key == self.INEQ_TOLERANCE:
                self.problem.ineq_tolerance = options[option_key]
                del options[option_key]
            elif option_key == self.EQ_TOLERANCE:
                self.problem.eq_tolerance = options[option_key]
                del options[option_key]
            elif options[option_key] is None:
                del options[option_key]

        try:
            self.opt_grammar.load_data(options)
        except InvalidDataException:
            raise ValueError("Invalid options for algorithm " + self.opt_grammar.name)
        for option_key in list(options.keys()):  # Copy keys on purpose
            lib_option_key = self.OPTIONS_MAP.get(option_key)
            # Overload with specific keys
            if lib_option_key is not None:
                options[lib_option_key] = options[option_key]
                if lib_option_key != option_key:
                    del options[option_key]

        return options

    def _check_ignored_options(self, options):
        """Check that the user did not passed options that do not exist for this driver.

        Raises a warning if it is the case
        :param options: options dict
        """
        for option_key in options:
            if not self.driver_has_option(option_key):
                msg = "Driver " + self.algo_name + " has no option " + option_key
                msg += ", option is ignored !"
                LOGGER.warning(msg)

    def execute(self, problem, algo_name=None, **options):
        """Executes the driver.

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
            if None, use self.algo_name
            which may have been set by the factory (Default value = None)
        :param options: the options dict for the algorithm
        """
        self.problem = problem
        if algo_name is not None:
            self.algo_name = algo_name
        if self.algo_name is None:
            raise ValueError(
                "Algorithm name must be either passed as "
                + "argument or set by the attribute self.algo_name"
            )
        self.__check_algorithm(self.algo_name, problem)
        self.init_options_grammar(self.algo_name)
        use_database = options.get(self.USE_DATABASE_OPTION, True)
        normalize = options.get(self.NORMALIZE_DESIGN_SPACE_OPTION, True)
        round_ints = options.get(self.ROUND_INTS_OPTION, True)

        self._check_ignored_options(options)
        options = self._get_options(**options)
        self.internal_algo_name = self.lib_dict[self.algo_name][self.INTERNAL_NAME]
        problem.check()
        problem.preprocess_functions(
            normalize=normalize, use_database=use_database, round_ints=round_ints
        )

        try:  # Term criteria such as max iter or max_time can be triggered in pre_run
            self._pre_run(problem, self.algo_name, **options)
            result = self._run(**options)
        except TerminationCriterion as error:
            result = self._termination_criterion_raised(error)
        self.finalize_iter_observer()
        self.problem.clear_listeners()
        self._post_run(problem, algo_name, result, **options)

        return result

    def _termination_criterion_raised(self, error):  # pylint: disable=W0613
        """Retrieve the best known iterate when max iter has been reached.

        :param error: the obtained error from the algorithm
        """
        if isinstance(error, TerminationCriterion):
            message = ""
            if isinstance(error, MaxIterReachedException):
                message = "Maximum number of iterations reached."
            elif isinstance(error, FunctionIsNan):
                message = "Function value or gradient or constraint is NaN, "
                message += "and problem.stop_if_nan is set to True."
            elif isinstance(error, DesvarIsNan):
                message = "Design variables are NaN."
            elif isinstance(error, XtolReached):
                message = "Successive iterates of the design variables "
                message += "are closer than xtol_rel or xtol_abs."
            elif isinstance(error, FtolReached):
                message = "Successive iterates of the objective function "
                message += "are closer than ftol_rel or ftol_abs."
            elif isinstance(error, MaxTimeReached):
                message = "Maximum time reached : {} seconds.".format(self._max_time)
            message += " GEMSEO Stopped the driver"
        else:
            message = error.args[0]

        result = self.get_optimum_from_database(message)
        return result

    def get_optimum_from_database(self, message=None, status=None):
        """Retrieves the optimum from the database and builds an optimization result
        object from it.

        :param message: Default value = None)
        :param status: Default value = None)
        """
        problem = self.problem
        if len(problem.database) == 0:
            return OptimizationResult(
                optimizer_name=self.algo_name,
                message=message,
                status=status,
                n_obj_call=0,
            )
        x_0 = problem.database.get_x_by_iter(0)
        # compute the best feasible or infeasible point
        f_opt, x_opt, is_feas, c_opt, c_opt_grad = problem.get_optimum()
        if f_opt is not None and not problem.minimize_objective:
            f_opt = -f_opt
        return OptimizationResult(
            x_0=x_0,
            x_opt=x_opt,
            f_opt=f_opt,
            optimizer_name=self.algo_name,
            message=message,
            status=status,
            n_obj_call=problem.objective.n_calls,
            is_feasible=is_feas,
            constraints_values=c_opt,
            constraints_grad=c_opt_grad,
        )

    def _get_options(self, **options):
        """Retrieves the options of the library To be overloaded by subclasses Used to
        define default values for options using keyword arguments.

        :param options: options of the driver
        """
        raise NotImplementedError()

    def _run(self, **options):
        """Runs the algorithm, to be overloaded by subclasses.

        :param options: the options dict for the algorithm
        """
        raise NotImplementedError()

    def __check_algorithm(self, algo_name, problem):
        """Checks that algorithm required by user is available and adapted to the
        problem. Set optimization library Set algorithm name according to optimization
        library requirements.

        :param algo_name : name of algorithm
        :type algo_name: str
        :param problem: optimization problem
        :type problem: OptimizationProblem
        """
        # Check that the algorithm is available
        if algo_name not in self.lib_dict:
            raise KeyError(
                "Requested optimization algorithm"
                + " %s is not in list of available algorithms %s"
                % (algo_name, list(self.lib_dict.keys()))
            )

        # Check that the algorithm is suited to the problem
        algo_dict = self.lib_dict[self.algo_name]
        if not self.is_algorithm_suited(algo_dict, problem):
            raise ValueError(
                "Algorithm {} is not adapted to the problem.".format(algo_name)
            )

    @staticmethod
    def _display_result(result):
        """Displays the optimization result.

        :param result: the result to display
        """
        LOGGER.info("Algorithm execution finished, result is:")
        LOGGER.info(result)

    @staticmethod
    def is_algorithm_suited(algo_dict, problem):
        """Checks if the algorithm is suited to the problem according to its algo dict.

        :param algo_dict: the algorithm characteristics
        :param problem: the opt_problem to be solved
        """
        raise NotImplementedError()

    def filter_adapted_algorithms(self, problem):
        """Filters the algorithms capable of solving the problem.

        :param problem: the opt_problem to be solved
        :returns: the list of adapted algorithms names
        """
        available = []
        for algo_name, algo_dict in self.lib_dict.items():
            if self.is_algorithm_suited(algo_dict, problem):
                available.append(algo_name)
        return available

    def is_algo_requires_grad(self, algo_name):
        """Returns True if the algorithm requires a gradient evaluation.

        :param algo_name: name of the algorithm
        """
        lib_alg = self.lib_dict.get(algo_name, None)
        if lib_alg is None:
            raise ValueError("Algorithm " + str(algo_name) + " is not available !")

        return lib_alg.get(self.REQUIRE_GRAD, False)

    def get_x0_and_bounds_vects(self, normalize_ds):
        """Gets x0, bounds, normalized or not depending on algo options, all as numpy
        arrays.

        :param normalize_ds: if True, normalizes all input vars
               that are not integers, according to design space
               normalization policy
        :returns: x, lower bounds, upper bounds
        """
        dspace = self.problem.design_space
        l_b = dspace.get_lower_bounds()
        u_b = dspace.get_upper_bounds()

        # remove normalization from options for algo
        if normalize_ds:
            norm_dict = dspace.normalize
            norm_array = dspace.dict_to_array(norm_dict)
            xvec = self.problem.get_x0_normalized()
            l_b = where(norm_array, zeros_like(xvec), l_b)
            u_b = where(norm_array, ones_like(xvec), u_b)
        else:
            xvec = self.problem.design_space.get_current_x()

        return xvec, l_b, u_b

    def ensure_bounds(self, orig_func, normalize=True):
        """Project the design vector onto the design space before execution.

        :param orig_func: the original function
        :param normalize: if True, use the normalized design space
        :returns: the wrapped function
        """

        def wrapped_func(x_vect):
            x_proj = self.problem.design_space.project_into_bounds(x_vect, normalize)
            return orig_func(x_proj)

        return wrapped_func



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A generic dataset to store data in memory.

This module implements the concept of dataset
which is a key element for machine learning, post-processing, data analysis, ...

A :class:`.Dataset` uses its attribute :attr:`.Dataset.data`
to store :math:`N` series of data
representing the values of :math:`p` multidimensional features
belonging to different groups of features.

This attribute :attr:`.Dataset.data` is a dictionary of 2D numpy arrays,
whose rows are the samples, a.k.a. series, realizations or entries,
and columns are the variables, a.k.a. parameters or features.
The keys of this dictionary are
either the names of the groups of variables
or the names of the variables.
Thus, a :class:`.Dataset` is not only defined by the raw data stored
but also by the names, the sizes and the groups of the different variables.

A :class:`.Dataset` can be set
either from a file (:meth:`.Dataset.set_from_file`)
or from a numpy arrays (:meth:`.Dataset.set_from_array`),
and can be enriched from a group of variables (:meth:`.Dataset.add_group`)
or from a single variable (:meth:`.Dataset.add_variable`).

An :class:`.AbstractFullCache` or an :class:`.OptimizationProblem`
can also be exported to a :class:`.Dataset`
using :meth:`.AbstractFullCache.export_to_dataset`
and :meth:`.OptimizationProblem.export_to_dataset` respectively.

From a :class:`.Dataset`,
we can easily access its length and data,
either as 2D array or as dictionaries indexed by the variables names.
We can get either the whole data,
or the data associated to a group or the data associated to a list of variables.
It is also possible to export the :class:`.Dataset`
to an :class:`.AbstractFullCache` or a pandas DataFrame.
"""
from __future__ import division, unicode_literals

import logging
import operator
from numbers import Number
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
from numpy import (
    array,
    concatenate,
    delete,
    genfromtxt,
    hstack,
    isnan,
    ndarray,
    unique,
    where,
)
from six import string_types

from gemseo.caches.cache_factory import CacheFactory
from gemseo.core.cache import AbstractFullCache
from gemseo.post.dataset.factory import DatasetPlotFactory
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.py23_compat import long
from gemseo.utils.string_tools import MultiLineString, pretty_repr

LOGGER = logging.getLogger(__name__)

LOGICAL_OPERATORS = {
    "<=": operator.le,
    "<": operator.lt,
    ">=": operator.ge,
    ">": operator.gt,
    "==": operator.eq,
    "!=": operator.ne,
}

ItemType = Union[
    int, str, List[int], List[str], Tuple[Union[int, List[int]], Union[str, List[str]]]
]
AllDataType = Union[
    Dict[str, Union[Dict[str, ndarray], ndarray]],
    Tuple[Union[ndarray, Dict[str, ndarray]], List[str], Dict[str, int]],
]


class Dataset(object):
    """A generic class to store data.

    Attributes:
        name (str): The name of the dataset.
        sizes (Dict[str,int]): The sizes of the variables.
        dimension (Dict[str,int]): The dimensions of the groups of variables.
        length (int): The length of the dataset.
        strings_encoding (Dict): The encoding structure,
            mapping the values of the string variables with integers;
            the keys are the names of the variables
            and the values are dictionaries
            whose keys are the components of the variables
            and the values are the integer values.
        metadata (Dict[str,Any]): The metadata
            used to store any kind of information that are not variables,
            e.g. the mesh associated with a multi-dimensional variable.
    """

    PARAMETER_GROUP = "parameters"
    DESIGN_GROUP = "design_parameters"
    FUNCTION_GROUP = "functions"
    INPUT_GROUP = "inputs"
    OUTPUT_GROUP = "outputs"
    GRADIENT_GROUP = "gradients"
    DEFAULT_GROUP = PARAMETER_GROUP
    DEFAULT_NAMES = {
        PARAMETER_GROUP: "x",
        DESIGN_GROUP: "dp",
        FUNCTION_GROUP: "func",
        INPUT_GROUP: "in",
        OUTPUT_GROUP: "out",
    }

    HDF5_CACHE = "HDF5Cache"
    MEMORY_FULL_CACHE = "MemoryFullCache"

    def __init__(
        self,
        name=None,  # type: Optional[str]
        by_group=True,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            name: The name of the dataset.
                If None, use the name of the class.
            by_group: If True, store the data by group.
                Otherwise, store them by variables.
        """
        self.name = name or self.__class__.__name__
        self._names = {}  # key = group, value = varnames
        self._groups = {}  # key = varname, value = group
        self.sizes = {}  # key = varname, value = varsize
        self._positions = {}
        self.dimension = {}  # key = group, value = groupsize
        self.length = 0
        self.data = {}
        self._group = by_group
        self.strings_encoding = None
        self._cached_inputs = []
        self._cached_outputs = []
        self.metadata = {}
        self.__row_names = None

    def remove(
        self,
        entries,  # type: Union[List[int],ndarray]
    ):  # type: (...) -> None
        """Remove entries.

        Args:
            entries: The entries to be removed,
                either indices
                or a boolean 1D array
                whose length is equal to the length of the dataset
                and elements to delete are coded True.
        """
        if isinstance(entries, ndarray):
            entries = self.find(entries)
        self.length -= len(entries)
        for name, value in list(self.data.items()):
            self.data[name] = delete(value, entries, 0)

    @staticmethod
    def find(
        comparison,  # type: ndarray
    ):  # type: (...) -> List[int]
        """Find the entries for which a comparison is satisfied.

        This search uses a boolean 1D array
        whose length is equal to the length of the dataset.

        Args:
            comparison: A boolean vector whose length is equal to the number of samples.

        Returns:
            The indices of the entries for which the comparison is satisfied.
        """
        return where(comparison)[0].tolist()

    def is_nan(self):  # type: (...) -> ndarray
        """Check if an entry contains NaN.

        Returns:
             Whether any entries is NaN or not.
        """
        return isnan(self.get_all_data(False)[0]).any(1)

    def compare(
        self,
        value_1,  # type: Union[str,float]
        logical_operator,  # type: str
        value_2,  # type: Union[str,float]
        component_1=0,  # type: int
        component_2=0,  # type: int
    ):  # type: (...) -> ndarray
        """Compare either a variable and a value or a variable and another variable.

        Args:
            value_1: The first value,
                either a variable name or a numeric value.
            logical_operator: The logical operator,
                either "==", "<", "<=", ">" or ">=".
            value_2: The second value,
                either a variable name or a numeric value.
            component_1: If value_1 is a variable name,
                component_1 corresponds to its component used in the comparison.
            component_2: If value_2 is a variable name,
                component_2 corresponds to its component used in the comparison.

        Returns:
             Whether the comparison is valid for the different entries.
        """
        if value_1 not in self.variables and value_2 not in self.variables:
            raise ValueError(
                "Either value_1 ({}) or value_2 ({}) "
                "must be a variable name from the list: {}".format(
                    value_1, value_2, self.variables
                )
            )
        if value_1 in self.variables:
            value_1 = self[value_1][value_1][:, component_1]
        if value_2 in self.variables:
            value_2 = self[value_2][value_2][:, component_2]
        try:
            result = LOGICAL_OPERATORS[logical_operator](value_1, value_2)
        except KeyError:
            raise ValueError(
                "{} is not a logical operator: "
                "use either '==', '<', '<=', '>' or '>='".format(logical_operator)
            )
        return result

    def _clean(self):  # type: (...) -> None
        """Remove all data from the dataset."""
        self._names = {}
        self._groups = {}
        self.sizes = {}
        self._positions = {}
        self.dimension = {}
        self.length = 0
        self.data = {}
        self.strings_encoding = None
        self._cached_inputs = []
        self._cached_outputs = []
        self.metadata = {}

    def is_group(
        self,
        name,  # type: str
    ):  # type: (...) -> bool
        """Check if a name is a group name.

        Args:
            name: A name of a group.

        Returns:
            Whether the name is a group name.
        """
        return name in self._names

    def is_variable(
        self,
        name,  # type: str
    ):  # type: (...) -> bool
        """Check if a name is a variable name.

        Args:
            name: A name of a variable.

        Returns:
            Whether the name is a variable name.
        """
        return name in self._groups

    def is_empty(self):  # type: (...) -> bool
        """Check if the dataset is empty.

        Returns:
            Whether the dataset is empty.
        """
        return self.n_samples == 0

    def get_names(
        self,
        group_name,  # type: str
    ):  # type: (...) -> List[str]
        """Get the names of the variables of a group.

        Args:
            group_name: The name of the group.

        Returns:
            The names of the variables of the group.
        """
        return self._names.get(group_name)

    def get_group(
        self,
        variable_name,  # type: str
    ):  # type: (...) -> str
        """Get the name of the group that contains a variable.

        Args:
            variable_name: The name of the variable.

        Returns:
            The group to which the variable belongs.
        """
        return self._groups.get(variable_name)

    @property
    def variables(self):  # type: (...) -> List[str]
        """The sorted names of the variables."""
        return sorted(self._groups.keys())

    @property
    def groups(self):  # type: (...) -> List[str]
        """The sorted names of the groups of variables."""
        return sorted(self._names.keys())

    def __str__(self):  # type: (...) -> str
        msg = MultiLineString()
        msg.add(self.name)
        msg.indent()
        msg.add("Number of samples: {}", self.n_samples)
        msg.add("Number of variables: {}", self.n_variables)
        msg.add("Variables names and sizes by group:")
        msg.indent()
        for group, varnames in sorted(self._names.items()):
            varnames = ["{} ({})".format(name, self.sizes[name]) for name in varnames]
            if varnames:
                msg.add("{}: {}", group, pretty_repr(varnames))
        total = sum(self.dimension.values())
        msg.dedent()
        msg.add("Number of dimensions (total = {}) by group:", total)
        msg.indent()
        for group, size in sorted(self.dimension.items()):
            msg.add("{}: {}", group, size)
        return str(msg)

    def __check_new_variable(
        self,
        variable,  # type: str
    ):  # type: (...) -> None
        """Check if a variable is defined.

        Args:
            variable: The name of the variable.
        """
        if self.is_variable(variable):
            raise ValueError("{} is already defined.".format(variable))
        if not isinstance(variable, string_types):
            raise TypeError("{} is not a string.".format(variable))

    def __check_new_group(self, group):
        """Check if a group is defined.

        Args:
            group: The name of the group.
        """
        if self.is_group(group):
            raise ValueError("{} is already defined.".format(group))
        if not isinstance(group, string_types):
            raise TypeError("{} is not a string.".format(group))

    def __check_length_consistency(
        self,
        length,  # type: int
    ):  # type: (...) -> None
        """Check if a length is consistent with the length of the dataset and set it.

        Args:
            length: A length to be tested.

        Raises:
            ValueError: If the tested length is different from the dataset one.
        """
        if self.length != 0 and length != self.length:
            raise ValueError(
                "The number of rows of data must be equal to the"
                " length of the dataset."
            )
        self.length = length

    def __check_data_consistency(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> None
        """Check that a data array is consistent.

        It must me a 2D numpy array with length equal to the dataset one.

        Raises:
            ValueError: If the data is not a 2D numpy array.
        """
        if not isinstance(data, ndarray) or len(data.shape) != 2:
            raise TypeError("data must be a 2D numpy array.")
        self.__check_length_consistency(data.shape[0])

    @staticmethod
    def __check_variables_format(
        variables,  # type: List[str]
    ):  # type: (...) -> None
        """Check that the names of the variables are well formatted.

        Args:
            variables: The names of the variables.
        """
        if not isinstance(variables, list):
            raise TypeError("variables must be a list of strings.")
        if any([not isinstance(name, string_types) for name in variables]):
            raise TypeError("variables must be a list of strings.")

    @staticmethod
    def __check_sizes_format(
        sizes,  # type: Dict[str,int]
        variables,  # type: Iterable[int],
        dimension,  # type: int
    ):  # type:(...) -> None
        """Check that the sizes of the variables are well specified.

        Args:
            sizes: The sizes of the variables.
            variables: The names of the variables.
            dimension: The data dimension.
        """

        def is_size(size):
            return isinstance(size, (int, long)) and size > 0

        if not isinstance(sizes, dict):
            raise TypeError("sizes must be a dictionary of positive integers.")
        if any([not is_size(sizes.get(name)) for name in variables]):
            raise TypeError("sizes must be a dictionary of positive integers.")
        total = sum([sizes[name] for name in variables])
        if total != dimension:
            raise ValueError(
                "The sum of variables sizes ({}) must be equal "
                "to the data dimension ({}).".format(total, dimension)
            )

    def __check_variables_sizes(
        self,
        variables,  # type: List[str]
        sizes,  # type: Dict[str,int]
        dimension,  # type: int
    ):  # type: (...) -> None
        """Check that the variables are well formatted.

        Args:
            sizes: The sizes of the variables.
            variables: The names of the variables.
            dimension: The data dimension.
        """
        if variables is not None:
            self.__check_variables_format(variables)
            if sizes is not None:
                self.__check_sizes_format(sizes, variables, dimension)

    def __get_default_group_variables(
        self,
        group,  # type: str
        dimension,  # type: int
        pattern=None,  # type: Optional[str]
    ):  # type: (...) -> Tuple[List[str], Dict[str,int],Dict[str,str]]
        """Create default names of the variables of a group.

        Args:
            group: The name of the group.
            dimension: The dimension of the group.
            pattern: A pattern to be used for the default name,
                e.g. 'x' will lead to 'x_0', 'x_1', ...
                If None,
                use :attr:`.Dataset.DEFAULT_NAMES`
                or the name of the group.

        Returns:
            The names, the sizes and the groups of the variables.
        """
        pattern = pattern or self.DEFAULT_NAMES.get(group) or group
        variables = ["{}_{}".format(pattern, index) for index in range(dimension)]
        sizes = {name: 1 for name in variables}
        groups = {name: group for name in variables}
        return variables, sizes, groups

    def __set_group_data(
        self,
        data,  # type: ndarray
        group,  # type: str
        variables,  # type: Iterable[str]
        sizes,  # type: Dict[str,int]
    ):  # type: (...) -> None
        """Set the data related to a group.

        Args:
            data: The data.
            group: The name of the group.
            variables: The names of the variables.
            sizes: the sizes of the variables.
        """
        if self._group:
            self.data[group] = data
        else:
            array_to_dict = DataConversion.array_to_dict
            self.data.update(array_to_dict(data, variables, sizes))

    def __set_variable_data(
        self,
        name,  # type: str
        data,  # type: ndarray
        group,  # type: str
    ):  # type: (...) -> None
        """Set the data related to a variable.

        Args:
            name: The name of the variable.
            data: The data.
            group: The name of the group.
        """
        if self._group:
            if not self.is_group(group):
                self.data[group] = data
            else:
                self.data[group] = hstack((self.data[group], data))
        else:
            self.data[name] = data

    def __set_group_properties(
        self,
        group,  # type: str
        variables,  # type: List[str]
        sizes,  # type: Dict[str,int],
        cache_as_input,  # type: bool
    ):  # type: (...) -> None
        """Set the properties related to a group.

        Args:
            group: The name of the group.
            variables: The names of the variables.
            sizes: The sizes of the variables.
            cache_as_input: If True,
                cache these data as inputs
                when the cache is exported to a cache.
        """
        self.sizes.update(sizes)
        self._groups.update({name: group for name in variables})
        self._names[group] = variables
        self.dimension[group] = sum([sizes[name] for name in variables])
        start = 0
        for name in variables:
            self._positions[name] = [start, start + self.sizes[name] - 1]
            if self._group:
                start += self.sizes[name]
            if cache_as_input:
                self._cached_inputs.append(name)
            else:
                self._cached_outputs.append(name)

    def __set_variable_properties(
        self,
        variable,  # type: str
        group,  # type: str
        size,  # type:int
        cache_as_input,  # type: bool
    ):  # type: (...) -> None
        """Set the properties related to a variable.

        Args:
            variable: The name of the variable.
            group: The name of the group.
            size: The size of the variable.
            cache_as_input: If True,
                cache these data as inputs
                when the cache is exported to a cache.
        """
        self.sizes[variable] = size
        self._groups[variable] = group
        if not self.is_group(group):
            self._names[group] = [variable]
            self.dimension[group] = self.sizes[variable]
            self._positions[variable] = [0, self.sizes[variable] - 1]
        else:
            self._names[group].append(variable)
            if self._group:
                ncols = self.dimension[group]
                self._positions[variable] = [ncols, ncols + self.sizes[variable] - 1]
            else:
                self._positions[variable] = [0, self.sizes[variable] - 1]
            self.dimension[group] += self.sizes[variable]
        if cache_as_input:
            self._cached_inputs.append(variable)
        else:
            self._cached_outputs.append(variable)

    def add_group(
        self,
        group,  # type: str
        data,  # type: ndarray
        variables=None,  # type: Optional[List[str]]
        sizes=None,  # type: Optional[Dict[str,int]]
        pattern=None,  # type: Optional[str]
        cache_as_input=True,  # type: bool
    ):  # type: (...) -> str
        """Add data related to a group.

        Args:
            group: The name of the group of data to be added.
            data: The data to be added.
            variables: The names of the variables.
                If None, use default names based on a pattern.
            sizes: The sizes of the variables.
                If None,
                assume that all the variables have a size equal to 1.
            pattern: The name of the variable to be used as a pattern
                when variables is None.
                If None,
                use the :attr:`.Dataset.DEFAULT_NAMES` for this group if it exists.
                Otherwise, use the group name.
            cache_as_input: If True,
                cache these data as inputs
                when the cache is exported to a cache.
        """
        self.__check_new_group(group)
        self.__check_data_consistency(data)
        self.__check_variables_sizes(variables, sizes, data.shape[1])
        if variables is None or sizes is None:
            variables, sizes, _ = self.__get_default_group_variables(
                group, data.shape[1], pattern
            )
        self.__set_group_data(data, group, variables, sizes)
        self.__set_group_properties(group, variables, sizes, cache_as_input)

    def add_variable(
        self,
        name,  # type: str
        data,  # type: ndarray
        group=DEFAULT_GROUP,  # type: str
        cache_as_input=True,  # type: bool
    ):  # type:(...) -> None
        """Add data related to a variable.

        Args:
            name: The name of the variable to be stored.
            data: The data to be stored.
            group: The name of the group related to this variable.
            cache_as_input: If True,
                cache these data as inputs
                when the cache is exported to a cache.
        """
        self.__check_new_variable(name)
        self.__check_data_consistency(data)
        self.__set_variable_data(name, data, group)
        self.__set_variable_properties(name, group, data.shape[1], cache_as_input)

    def __get_strings_encoding(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> Tuple[ndarray,Dict[str,Dict[int,int]]]
        """Encode string data.

        Args:
            data: The data to be encoded.

        Returns:
            The encoded data and the encoding structure.
        """
        self.strings_encoding = {name: {} for name in self._groups}
        encoding = {}
        if str(data.dtype).startswith(("|S", "<U")):
            data, encoding = self.__force_array_to_float(data)
        return data, encoding

    def set_from_array(
        self,
        data,  # type: ndarray
        variables=None,  # type: Optional[List[str]]
        sizes=None,  # type: Optional[Dict[str,int]]
        groups=None,  # type: Optional[Dict[str,str]]
        default_name=None,  # type: Optional[str]
    ):  # type: (...) -> None
        """Set the dataset from an array.

        Args:
            data: The data to be stored.
            variables: The names of the variables.
                If None,
                use one default name per column of the array
                based on the pattern 'default_name'.
            sizes: The sizes of the variables.
                If None,
                assume that all the variables have a size equal to 1.
            groups: The groups of the variables.
                If None,
                use :attr:`.Dataset.DEFAULT_GROUP` for all the variables.
            default_name: The name of the variable to be used as a pattern
                when variables is None.
                If None,
                use the :attr:`.Dataset.DEFAULT_NAMES` for this group if it exists.
                Otherwise, use the group name.
        """
        self._clean()
        self.__check_data_consistency(data)
        if variables is None:
            group = self.DEFAULT_GROUP
            get = self.__get_default_group_variables
            variables, sizes, groups = get(group, data.shape[1], default_name)
        else:
            self.__check_variables_format(variables)
        if sizes is None:
            sizes = {name: 1 for name in variables}
        self.__check_sizes_format(sizes, variables, data.shape[1])
        if groups is None:
            groups = {name: self.DEFAULT_GROUP for name in variables}
        self.__check_groups_format(groups, variables)
        self.__set_data_properties(variables, sizes, groups)
        data, encoding = self.__get_strings_encoding(data)
        self.__set_data(data, variables, encoding)

    def __check_groups_format(
        self,
        groups,  # type: Dict[str,str]
        variables,  # type: Iterable[str]
    ):  # type: (...) -> None
        """Check the format of groups and update it if necessary.

        Args:
            groups: The names of the groups of the variables.
            variables: The names of the variables.
        """
        if not isinstance(groups, dict):
            raise TypeError(
                "groups must be a dictionary indexed by variables"
                "names whose values are strings."
            )
        for name in variables:
            if groups.get(name) is None:
                groups.update({name: self.DEFAULT_GROUP})
            elif not isinstance(groups[name], string_types):
                raise TypeError(
                    "groups must be a dictionary indexed "
                    "by variables names whose values are strings."
                )

    def __set_data(
        self,
        data,  # type: ndarray
        variables,  # type: Iterable[str]
        encoding,  # type: Dict[str,Dict[int,int]]
    ):  # type: (...) -> None
        """Set data.

        Args:
            data: The data to be stored.
            variables: The names of the variables.
            encoding: An encoding structure
                mapping the values of the string variables with integers;
                the keys are the names of the variables
                and the values are dictionaries
                whose keys are the components of the variables
                and the values are the integer values.
        """
        indices = {group: [] for group in self._names}
        start = 0
        for name in variables:
            end = start + self.sizes[name] - 1
            name_indices = list(range(start, end + 1))
            start = end + 1
            indices[self._groups[name]] += name_indices
            for key in encoding:
                if key in name_indices:
                    index = name_indices.index(key)
                    self.strings_encoding[name][index] = encoding[key]
            if not self._group:
                self.data[name] = data[:, name_indices]
        if self._group:
            for group in self._names:
                self.data[group] = data[:, indices[group]]

    def _set_variables_positions(self):
        """Set variables positions."""
        for varnames in self._names.values():
            start = 0
            for varname in varnames:
                self._positions[varname] = [start, start + self.sizes[varname] - 1]
                if self._group:
                    start += self.sizes[varname]
                else:
                    start = 0

    def __set_data_properties(
        self,
        variables,  # type: Iterable[str]
        sizes,  # type: Mapping[str,int]
        groups,  # type: Mapping[str,str]
    ):  # type: (...) -> None
        """Set the properties for the whole data.

        Args:
            variables: The names of the variables.
            sizes: The sizes of the variables.
            groups: The groups of the variables.
        """
        for name in variables:
            if not self.is_group(groups[name]):
                self._names[groups[name]] = [name]
            else:
                self._names[groups[name]].append(name)
            self.sizes[name] = sizes[name]
            self._groups[name] = groups[name]
        for group, names in self._names.items():
            self.dimension[group] = sum([self.sizes[name] for name in names])
            if group == self.OUTPUT_GROUP:
                self._cached_outputs += names
            else:
                self._cached_inputs += names
        self._set_variables_positions()

    @staticmethod
    def __force_array_to_float(
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Force a ndarray type to float.

        Args:
            data: The data to be casted.

        Returns:
            The data with float type.
        """

        def __is_not_float(
            obj,  # type: Any
        ):  # type: (...) -> bool
            """Return True if an object cannot be cast to float.

            Args:
                obj: The object to test.

            Returns:
                Whether the object can be converted to float.
            """
            try:
                float(obj)
                return False
            except ValueError:
                return True

        first_row = data[0, :]
        str_indices = [
            index for index, element in enumerate(first_row) if __is_not_float(element)
        ]

        strings_encoding = {}
        for index in range(data.shape[1]):
            if index in str_indices:
                column = list(data[:, index])
                encoding = dict(enumerate(unique(column)))
                strings_encoding[index] = encoding
                data[:, index] = unique(column, return_inverse=True)[1]

        data = data.astype("float")
        return data, strings_encoding

    def set_from_file(
        self,
        filename,  # type: str
        variables=None,  # type: Optional[List[str]]
        sizes=None,  # type: Optional[Dict[str,int]]
        groups=None,  # type: Optional[Dict[str,str]]
        delimiter=",",  # type:str
        header=True,  # type: bool
    ):  # type: (...) -> None
        """Set the dataset from a file.

        Args:
            filename: The name of the file containing the data.
            variables: The names of the variables.
                If None and `header` is True,
                read the names from the first line of the file.
                If None and `header` is False,
                use default names
                based on the patterns the :attr:`.Dataset.DEFAULT_NAMES`
                associated with the different groups.
            sizes: The sizes of the variables.
                If None,
                assume that all the variables have a size equal to 1.
            groups: The groups of the variables.
                If None,
                use :attr:`.Dataset.DEFAULT_GROUP` for all the variables.
            delimiter: The field delimiter.
            header: If True,
                read the names of the variables on the first line of the file.
        """
        self._clean()
        data = genfromtxt(filename, delimiter=delimiter, dtype="str")
        if header:
            if variables is None:
                variables = data[0, :].tolist()
            start_read = 1
        else:
            start_read = 0
        self.set_from_array(data[start_read:, :], variables, sizes, groups)

    def set_metadata(
        self,
        name,  # type: str
        value,  # type: Any
    ):  # type: (...) -> None
        """Set a metadata attribute.

        Args:
            name: The name of the metadata attribute.
            value: The value of the metadata attribute.
        """
        self.metadata[name] = value

    @property
    def columns_names(self):  # type: (...) -> List[Union[str,Tuple[str,str,str]]]
        """The names of the columns of the dataset."""
        return self._get_columns_names()

    def _get_columns_names(
        self,
        as_list=False,  # type: bool
        start=0,  # type: int
    ):  # type: (...) -> List[Union[str,Tuple[str,str,str]]]
        """Return the names of the columns of the dataset.

        If dim(x)=1,
        its column name is 'x'
        while if dim(y)=2,
        its columns names are either 'x_0' and 'x_1'
        or [group_name, 'x', '0'] and [group_name, 'x', '1'].

        Args:
            as_list: If True, return the name as a tuple.
                otherwise, return the name as a string.
            start: The first index for the components of a variable.
                E.g. with '0': 'x_0', 'x_1', ...

        Returns:
            The names of the columns of the data.
        """
        column_names = []
        for group, names in self._names.items():
            for name in names:
                if as_list:
                    column_names += [
                        [group, name, str(index + start)]
                        for index in range(self.sizes[name])
                    ]
                else:
                    if self.sizes[name] == 1:
                        column_names += [name]
                    else:
                        column_names += [
                            "{}_{}".format(name, index + start)
                            for index in range(self.sizes[name])
                        ]
        return column_names

    def get_data_by_group(
        self,
        group,  # type: str
        as_dict=False,  # type: bool
    ):  # type: (...) -> Union[ndarray, Dict[str,ndarray]]
        """Get the data for a specific group name.

        Args:
            group: The name of the group.
            as_dict: If True, return values as dictionary.

        Returns:
            The data related to the group.
        """
        if not self.is_group(group):
            raise ValueError("{} is not an available group.".format(group))
        if group in self.data:
            data = self.data[group]
            if as_dict:
                data = DataConversion.array_to_dict(
                    self.data[group], self._names[group], self.sizes
                )
        else:
            data = {name: self.data[name] for name in self._names[group]}
            if not as_dict:
                data = DataConversion.dict_to_array(data, self._names[group])
        return data

    def get_data_by_names(
        self,
        names,  # type: Union[str,Iterable[str]]
        as_dict=True,  # type: bool
    ):  # type: (...) -> Union[ndarray, Dict[str,ndarray]]
        """Get the data for specific names of variables.

        Args:
            names: The names of the variables.
            as_dict: If True, return values as dictionary.

        Returns:
            The data related to the variables.
        """
        if isinstance(names, string_types):
            names = [names]
        dict_to_array = DataConversion.dict_to_array
        if not self._group:
            data = {name: self.data.get(name) for name in names}
        else:
            data = {}

            for name in names:
                indices = list(
                    range(self._positions[name][0], self._positions[name][1] + 1)
                )
                data[name] = self.data[self._groups[name]][:, indices]
        if not as_dict:
            data = dict_to_array(data, names)
        return data

    def get_all_data(self, by_group=True, as_dict=False):  # type: (...) -> AllDataType
        """Get all the data stored in the dataset.

        The data can be returned
        either as a dictionary indexed by the names of the variables,
        or as an array concatenating them,
        accompanied with the names and sizes of the variables.

        The data can also classified by groups of variables.

        Args:
            by_group: If True, sort the data by group.
            as_dict: If True, return the data as a dictionary.

        Returns:
            All the data stored in the dataset.
        """
        if by_group:
            data = {
                group: self.get_data_by_group(group, as_dict) for group in self._names
            }
            if not as_dict:
                data = (data, self._names, self.sizes)
        else:
            if not as_dict:
                data = concatenate(
                    tuple([self.get_data_by_group(group) for group in self._names]), 1
                )
                names = [
                    item for sublist in list(self._names.values()) for item in sublist
                ]
                data = (data, names, self.sizes)
            else:
                data = {}
                for group in self._names:
                    data.update(self.get_data_by_group(group, as_dict))
        return data

    @property
    def n_variables(self):  # type: (...) -> int
        """The number of variables."""
        return len(self._groups)

    def n_variables_by_group(
        self,
        group,  # type: str
    ):  # type: (...) -> int
        """The number of variables for a group.

        Args:
            group: The name of a group.

        Returns:
            The group dimension.
        """
        return len(self._names[group])

    @property
    def n_samples(self):  # type: (...) -> int
        """The number of samples."""
        return self.length

    def __len__(self):  # type: (...) -> int
        """The length of the dataset."""
        return self.length

    def __bool__(self):  # type: (...) -> bool
        """True is the dataset is not empty."""
        return not self.is_empty()

    def export_to_dataframe(
        self,
        copy=True,  # type: bool
    ):  # type: (...) -> DataFrame
        """Export the dataset to a pandas Dataframe.

        Args:
            copy: If True, copy data.
                Otherwise, use reference.

        Returns:
            A pandas DataFrame containing the dataset.
        """
        from pandas import DataFrame

        row1 = []
        row2 = []
        row3 = []
        for column in self._get_columns_names(True):
            row1.append(column[0])
            row2.append(column[1])
            row3.append(column[2])
        columns = [array(row1), array(row2), array(row3)]
        data = self.get_all_data(False, False)
        dataframe = DataFrame(data[0], columns=columns, copy=copy)
        dataframe.index = self.row_names
        return dataframe

    def export_to_cache(
        self,
        inputs=None,  # type: Optional[Iterable[str]]
        outputs=None,  # type: Optional[Iterable[str]]
        cache_type=MEMORY_FULL_CACHE,  # type: str
        cache_hdf_file=None,  # type: Optional[str]
        cache_hdf_node_name=None,  # type: Optional[str]
        **options
    ):  # type: (...) -> AbstractFullCache
        """Export the dataset to a cache.

        Args:
            inputs: The names of the inputs to cache.
                If None, use all inputs.
            outputs: The names of the outputs to cache.
                If None, use all outputs.
            cache_type: The type of cache to use.
            cache_hdf_file: The name of the HDF file to store the data.
                Required if the type of the cache is 'HDF5Cache'.
            cache_hdf_node_name: The name of the HDF node to store the discipline.
                If None, use the name of the dataset.

        Returns:
            A cache containing the dataset.
        """
        if inputs is None:
            inputs = self._cached_inputs
        if outputs is None:
            outputs = self._cached_outputs
        create_cache = CacheFactory().create
        cache_hdf_node_name = cache_hdf_node_name or self.name
        if cache_type == self.HDF5_CACHE:
            cache = create_cache(
                cache_type,
                hdf_file_path=cache_hdf_file,
                hdf_node_path=cache_hdf_node_name,
                **options
            )
        else:
            cache = create_cache(cache_type, **options)
        for sample in range(len(self)):
            in_values = {name: self[(sample, name)][name] for name in inputs}
            out_values = {name: self[(sample, name)][name] for name in outputs}
            cache.cache_outputs(in_values, inputs, out_values, outputs)
        return cache

    def get_available_plots(self):  # type: (...) -> List[str]
        """Return the available plot methods."""
        return DatasetPlotFactory().plots

    def plot(
        self,
        name,  # type:str
        show=True,  # type:bool
        save=False,  # type:bool
        **options
    ):  # type: (...) -> None
        """Plot the dataset from a :class:`DatasetPlot`.

        See :meth:`.Dataset.get_available_plots`

        Args:
            name: The name of the post-processing,
                which is the name of a class inheriting from :class:`DatasetPlot`.
            show: If True, display the figure.
            save: If True, save the figure.
            options: The options for the post-processing.
        """
        options.update({"show": show, "save": save})
        post = DatasetPlotFactory().create(name, dataset=self)
        post.execute(**options)
        return post

    def __getitem(
        self,
        indices,  # type: Sequence[int],
        names,  # type: Iterable[str],
    ):  # type: (...) -> Dict[str,ndarray]
        """Get the items associated with sample indices and variables names.

        Args:
            indices: The samples indices.

        Returns:
            The data related to the samples and variables names.
        """
        for index in indices:
            is_lower = index < -1
            is_greater = index >= self.length
            is_int = isinstance(index, int)
            if index == self.n_samples:
                raise IndexError
            if is_lower or is_greater or not is_int:
                raise ValueError("{} is not a sample index.".format(index))
        for name in names:
            if name not in self._groups:
                raise ValueError("'{}' is not a variable name.".format(name))
        item = self.get_data_by_names(names)
        if len(indices) == 1:
            indices = indices[0]
        item = {name: value[indices, :] for name, value in list(item.items())}
        return item

    def __getitem__(
        self,
        key,  # type: ItemType
    ):  # type: (...) -> Dict[str,ndarray]
        indices = list(range(0, self.length))
        type_error_msg = (
            "You can get items from a dataset in one of the following ways: "
            "dataset[3] for the 4th sample, "
            "dataset['x'] for the variable 'x', "
            "dataset[['x','y']] for the variables 'x' and 'y', "
            "dataset[[0,3]] for the 1st and 4th samples, "
            "dataset[(1,'x')] for the variable 'x' of the 2nd sample, "
            "dataset[(1,['x','y'])] for the variables 'x' and 'y' of the 2nd sample, "
            "dataset[([0,3],'x')] for the variable 'x' of the 1st and 4th samples, "
            "dataset[([0,3],['x','y'])] for the variables 'x' and 'y' "
            "of the 1st and 4th samples, "
        )

        def getitem_list(index):
            """Get the item when the index is a list."""
            if all(isinstance(elem, string_types) for elem in index):
                item = self.__getitem(indices, index)
            elif all(isinstance(elem, Number) for elem in index):
                item = self.__getitem(index, self.variables)
            else:
                raise TypeError(type_error_msg)
            return item

        def getitem_tuple(tpl):
            """Get the item when the index is a tuple."""
            if isinstance(tpl[0], Number):
                indices = [tpl[0]]
            elif isinstance(tpl[0], slice):
                indices = list(range(tpl[0].start, tpl[0].stop, tpl[0].step or 1))
            elif isinstance(tpl[0], list):
                if all(isinstance(elem, Number) for elem in tpl[0]):
                    indices = tpl[0]
                else:
                    raise TypeError(type_error_msg)
            else:
                raise TypeError(type_error_msg)
            if isinstance(tpl[1], string_types):
                names = [tpl[1]]
            elif isinstance(tpl[1], list):
                if all(isinstance(elem, string_types) for elem in tpl[1]):
                    names = tpl[1]
                else:
                    raise TypeError(type_error_msg)
            else:
                raise TypeError(type_error_msg)
            return self.__getitem(indices, names)

        if isinstance(key, Number):
            item = self.__getitem([key], self.variables)
        elif isinstance(key, string_types):
            item = self.__getitem(indices, [key])
        elif isinstance(key, list):
            item = getitem_list(key)
        elif isinstance(key, slice):
            item = getitem_list(list(range(key.start, key.stop, key.step or 1)))
        elif isinstance(key, tuple) and len(key) == 2:
            item = getitem_tuple(key)
        else:
            raise TypeError(type_error_msg)
        return item

    @property
    def row_names(self):  # type: (...) -> List[str]
        """The names of the rows."""

        return self.__row_names or [str(val) for val in range(len(self))]

    @row_names.setter
    def row_names(
        self,
        names,  # type: List[str]
    ):  # type: (...) -> None
        self.__row_names = names

    def get_normalized_dataset(
        self,
        excluded_variables=None,  # type: Optional[Sequence[str]]
        excluded_groups=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> Dataset
        """Get a normalized copy of the dataset.

        Args:
            excluded_variables: The names of the variables not to be normalized.
                If None, normalize all the variables.
            excluded_groups: The names of the groups not to be normalized.
                If None, normalize all the groups.

        Returns:
            A normalized dataset.
        """
        excluded_groups = excluded_groups or []
        excluded_variables = excluded_variables or []

        dataset = Dataset(self.name, self._group)
        for group, names in self._names.items():
            normalize_group = group not in excluded_groups

            for name in names:
                normalize_name = name not in excluded_variables
                data = self.get_data_by_names(name, False)
                if normalize_group and normalize_name:
                    data = (data - np.min(data, 0)) / (
                        np.max(data, 0) - np.min(data, 0)
                    )

                dataset.add_variable(name, data, group, name in self._cached_inputs)

        return dataset



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation
#               and/or initial documentation
#        :author: Sobieski, Agte, and Sandusky
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Damien Guenot
#        :author: Francois Gallard
# From NASA/TM-1998-208715
# Bi-Level Integrated System Synthesis (BLISS)
# Sobieski, Agte, and Sandusky
"""
SSBJ Propulsion computations
****************************
"""
from __future__ import division, unicode_literals

import logging
from math import pi

from numpy import append, array, atleast_2d, zeros

LOGGER = logging.getLogger(__name__)
DEG_TO_RAD = pi / 180.0


class SobieskiPropulsion(object):
    """Class defining propulsion analysis for Sobieski problem and related method to the
    propulsion problem such as disciplines computation, constraints, reference
    optimum."""

    DTYPE_COMPLEX = "complex128"
    DTYPE_DOUBLE = "float64"

    ESF_UPPER_LIMIT = 1.5
    ESF_LOWER_LIMIT = 0.5
    TEMPERATURE_LIMIT = 1.02

    def __init__(self, sobieski_base):
        """Constructor."""
        self.base = sobieski_base
        self.constants = self.base.default_constants()
        (
            self.x_initial,
            self.tc_initial,
            self.half_span_initial,
            self.aero_center_initial,
            self.cf_initial,
            self.mach_initial,
            self.h_initial,
            self.throttle_initial,
            self.lift_initial,
            self.twist_initial,
            self.esf_initial,
        ) = self.base.get_initial_values()
        self.dtype = self.base.dtype
        self.math = self.base.math
        # Surface fit to engine deck with least square method
        # Polynomial coefficients for SFC computation
        self.sfc_coeff = array(
            [
                1.13238425638512,
                1.53436586044561,
                -0.00003295564466,
                -0.00016378694115,
                -0.31623315541888,
                0.00000410691343,
                -0.00005248000590,
                -0.00000000008574,
                0.00000000190214,
                0.00000001059951,
            ],
            dtype=self.dtype,
        )
        # Polynomial coefficients for throttle constraint
        self.thua_coeff = array(
            [
                11483.7822254806,
                10856.2163466548,
                -0.5080237941,
                3200.157926969,
                -0.1466251679,
                0.0000068572,
            ],
            dtype=self.dtype,
        )
        self.throttle_coeff = self.dtype(16168.6)

    def __set_coeff_temp(self, x_shared, x_3):
        """Prepare settings of polynomial function for temperature.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :param x_3: local design variables
        :type x_3: numpy array
        :returns:  s_initial, s_new, flag, bound
        :rtype: numpy array
        """
        s_initial = array(
            [self.mach_initial, self.h_initial, self.throttle_initial], dtype=self.dtype
        )
        s_new = array([x_shared[2], x_shared[1], x_3[0]], dtype=self.dtype)
        flag = array([2, 4, 2], dtype=self.dtype)
        bound = array([0.25, 0.25, 0.25], dtype=self.dtype)
        return s_initial, s_new, flag, bound

    def compute_dim_throttle(self, adim_throttle):
        """Compute a dimensioned value of throttle from adim value.

        :param adim_throttle: local design vector
        :type adim_throttle: numpy array
        :returns: throttle
        :rtype: numpy array
        """
        return adim_throttle * self.throttle_coeff

    def compute_sfc(self, x_shared, adim_throttle):
        """Compute Specific Fuel Consumption (SFC) from global design variables (M, h,)
        and local design variable (throttle) by polynomial function.

        :param x_shared: global design vector
        :type x_shared: numpy array
        :param adim_throttle: local design vector
        :type adim_throttle: numpy array
        :returns: sfc
        :rtype: numpy array
        """
        throttle = self.compute_dim_throttle(adim_throttle)
        sfc = (
            self.sfc_coeff[0]
            + self.sfc_coeff[1] * x_shared[2]
            + self.sfc_coeff[2] * x_shared[1]
            + self.sfc_coeff[3] * throttle
            + self.sfc_coeff[4] * x_shared[2] ** 2
            + 2 * x_shared[1] * x_shared[2] * self.sfc_coeff[5]
            + 2 * throttle * x_shared[2] * self.sfc_coeff[6]
            + self.sfc_coeff[7] * x_shared[1] ** 2
            + 2 * throttle * x_shared[1] * self.sfc_coeff[8]
            + self.sfc_coeff[9] * throttle ** 2
        )
        return sfc

    def compute_throttle_ua(self, x_shared):
        """Compute throttle upper limit from global design variables (M, h,)

        :param x_shared: global design vector
        :type x_shared: numpy array
        :returns: throttle_uA
        :rtype: numpy array
        """
        throttle_ua = (
            self.thua_coeff[0]
            + self.thua_coeff[1] * x_shared[2]
            + self.thua_coeff[2] * x_shared[1]
            + self.thua_coeff[3] * x_shared[2] * x_shared[2]
            + 2 * self.thua_coeff[4] * x_shared[2] * x_shared[1]
            + self.thua_coeff[5] * x_shared[1] * x_shared[1]
        )
        return throttle_ua

    def compute_throttle_constraint(self, x_shared, adim_throttle):
        """Compute throttle constraint (M, h,) and local design variable (throttle) by
        polynomial function.

        :param x_shared: global design vector
        :type x_shared: numpy array
        :param adim_throttle: local design vector
        :type adim_throttle: numpy array
        :returns: throttle constraint as throttle / throttle_ua - 1.0
        :rtype: numpy array
        """
        throttle = self.compute_dim_throttle(adim_throttle)
        throttle_ua = self.compute_throttle_ua(x_shared)
        throttle_constraint = throttle / throttle_ua - 1.0  # throttle setting
        return throttle_constraint

    def compute_dthrconst_dthrottle(self, x_shared):
        """Compute derivative of throttle constraint wrt throttle.

        :param x_shared: global design vector
        :type x_shared: numpy array
        :returns: dthrottle_constraint_dthrottle
        :rtype: numpy array
        """
        throttle_ua = self.compute_throttle_ua(x_shared)
        return self.throttle_coeff / throttle_ua

    def compute_dthrcons_dh(self, x_shared, adim_throttle):
        """Compute derivative of throttle constraint wrt altitude.

        :param x_shared: global design vector
        :type x_shared: numpy array
        :param adim_throttle: local design vector
        :type adim_throttle: numpy array
        :returns: dthrottle_constraint_dh
        :rtype: numpy array
        """
        throttle = self.compute_dim_throttle(adim_throttle)
        throttle_ua = self.compute_throttle_ua(x_shared)
        dthrottle_ua_dh = (
            self.thua_coeff[2]
            + 2 * self.thua_coeff[4] * x_shared[2]
            + 2.0 * self.thua_coeff[5] * x_shared[1]
        )
        return -throttle * dthrottle_ua_dh / (throttle_ua * throttle_ua)

    def compute_dthrconst_dmach(self, x_shared, adim_throttle):
        """Compute derivative of throttle constraint wrt Mach number.

        :param x_shared: global design vector
        :type x_shared: numpy array
        :param adim_throttle: local design vector
        :type adim_throttle: numpy array
        :returns: dthrottle_constraint_dmach
        :rtype: numpy array
        """
        throttle = self.compute_dim_throttle(adim_throttle)
        throttle_ua = self.compute_throttle_ua(x_shared)
        dthrottle_ua_dmach = (
            self.thua_coeff[1]
            + 2 * self.thua_coeff[3] * x_shared[2]
            + 2.0 * self.thua_coeff[4] * x_shared[1]
        )
        return -throttle * dthrottle_ua_dmach / (throttle_ua * throttle_ua)

    def compute_esf(self, drag, adim_throttle):
        """Compute engine scale factor.

        :param drag: drag
        :type drag: float
        :param adim_throttle: throttle (not dimensioned)
        :type adim_throttle: float
        :returns: Engine Scale Factor
        :rtype: numpy array
        """
        return drag / (3.0 * self.compute_dim_throttle(adim_throttle))

    def compute_desf_ddrag(self, adim_throttle):
        """Compute derivative of ESF wrt aero drag.

        :param adim_throttle: local design variables
        :type adim_throttle: numpy array
        :returns: derivative of ESF wrt drag
        :rtype: numpy array
        """
        return 1.0 / (3 * self.compute_dim_throttle(adim_throttle))

    def compute_desf_dthrottle(self, drag, adim_throttle):
        """Compute derivative of ESF wrt aero drag.

        :param drag: coupling design variables
        :type drag: numpy array
        :param adim_throttle: local design variables
        :type adim_throttle: numpy array
        :returns: derivative of ESF wrt drag
        :rtype: numpy array
        """
        throttle = self.compute_dim_throttle(adim_throttle)
        return -self.throttle_coeff * drag / (3.0 * throttle ** 2)

    def compute_temp(self, x_shared, x_3):
        """Compute engine temperature.

        :param x_3: local design vector
        :type x_3: numpy array
        :param x_shared: global design vector
        :type x_shared: numpy array
        :returns: engine temperature
        :rtype: numpy array
        """
        s_initial, s_new, flag, bound = self.__set_coeff_temp(x_shared, x_3)
        return self.base.poly_approx(s_initial, s_new, flag, bound)

    def compute_engine_weight(self, esf):
        """Compute engine weight.

        :param esf: engine scale factor
        :returns: engine weight
        :rtype: numpy array
        """
        return self.constants[3] * (esf ** 1.05) * 3

    def blackbox_propulsion(self, x_shared, y_23, x_3, true_cstr=False):
        """This function calculates fuel comsumption, engine weight and engine scale
        factor.

        :param x_shared: shared design variable vector:

            - x_shared[0]: thickness/chord ratio
            - x_shared[1]: altitude
            - x_shared[2]: Mach
            - x_shared[3]: aspect ratio
            - x_shared[4]: wing sweep
            - x_shared[5]: wing surface area

        :type x_shared: numpy array
        :param y_23: shared variables coming from blackbox_aerodynamics (drag)
        :type y_23: numpy array
        :param x_3: power/propulsion design variable (throttle setting)
        :type x_3: numpy array
        :param true_cstr: Default value = False)
        :returns: y_3, y_34, y_31, y_32, g_3:

            - y_3: output variables for propulsion analysis

                - y_3[0]: SFC
                - y_3[1]: engine weight
                - y_3[2]: engine scale factor

            - y_34: shared variable for BlackBoxMission (SFC)
            - y_31: shared variable for blackbox_structure (engine weight)
            - y_32: shared variable for blackbox_aerodynamics (esf)
            - g_3: propulsion constraints

                - g_3[0]: engine scale factor constraint
                - g_3[1]: engine temperature
                - g_3[2]: throttle setting constraint

        :rtype: numpy array, numpy array, numpy array, numpy array, numpy array
        """
        y_3 = zeros(3, dtype=self.dtype)
        g_3 = zeros(3, dtype=self.dtype)
        y_31 = zeros(1, dtype=self.dtype)
        y_32 = zeros(1, dtype=self.dtype)
        y_34 = zeros(1, dtype=self.dtype)

        esf = self.compute_esf(y_23[0], x_3[0])
        y_3[2] = esf
        y_3[1] = self.compute_engine_weight(esf)
        y_3[0] = self.compute_sfc(x_shared, x_3[0])

        y_31[0] = y_3[1]
        y_34[0] = y_3[0]
        y_32[0] = y_3[2]

        # THIS SECTION COMPUTES SFC, esf, AND ENGINE WEIGHT
        g_3[0] = y_3[2]  # engine scale factor

        # engine temperature
        temp = self.compute_temp(x_shared, x_3)
        g_3[1] = temp

        g_3[2] = self.compute_throttle_constraint(x_shared, x_3[0])

        if not true_cstr:
            g_3 = append(
                g_3[0] - self.ESF_UPPER_LIMIT,
                (
                    self.ESF_LOWER_LIMIT - g_3[0],
                    g_3[2],
                    g_3[1] - self.TEMPERATURE_LIMIT,
                ),
            )
        return y_3, y_34, y_31, y_32, g_3

    def compute_dengineweight_dvar(self, esf, desf_dx):
        """Computes derivative of engine weight wrt to a variable(drag or throttle)

        :param esf: ESF
        :type esf: float
        :param desf_dx: partial derivative of ESF wrt to input variable
        :type desf_dx: numpy array
        :returns: derivative of engine weight wrt a variable (drag or throttle)
        :rtype: numpy array
        """
        return 3 * self.constants[3] * 1.05 * desf_dx * esf ** 0.05

    def compute_dsfc_dthrottle(self, x_shared, adim_throttle):
        """Compute derivative of sfc constraint wrt throttle.

        :param x_shared: global design vector
        :type x_shared: numpy array
        :param adim_throttle: local design vector
        :type adim_throttle: numpy array
        :returns: dsfc_dthrottle
        :rtype: numpy array
        """
        throttle = self.compute_dim_throttle(adim_throttle)
        dsfc_dthrottle = (
            self.sfc_coeff[3]
            + 2 * x_shared[2] * self.sfc_coeff[6]
            + 2 * x_shared[1] * self.sfc_coeff[8]
            + 2 * self.sfc_coeff[9] * throttle
        )
        dsfc_dthrottle *= self.throttle_coeff
        return dsfc_dthrottle

    def compute_dsfc_dh(self, x_shared, adim_throttle):
        """Compute derivative of sfc constraint wrt altitude.

        :param x_shared: global design vector
        :type x_shared: numpy array
        :param adim_throttle: local design vector
        :type adim_throttle: numpy array
        :returns: dsfc_dh
        :rtype: numpy array
        """
        throttle = self.compute_dim_throttle(adim_throttle)
        dsfc_dh = (
            self.sfc_coeff[2]
            + 2 * x_shared[2] * self.sfc_coeff[5]
            + +2 * self.sfc_coeff[7] * x_shared[1]
            + 2 * throttle * self.sfc_coeff[8]
        )
        return dsfc_dh

    def compute_dsfc_dmach(self, x_shared, adim_throttle):
        """Compute derivative of sfc constraint wrt Mach number.

        :param x_shared: global design vector
        :type x_shared: numpy array
        :param adim_throttle: local design vector
        :type adim_throttle: numpy array
        :returns: dsfc_dmach
        :rtype: numpy array
        """
        throttle = self.compute_dim_throttle(adim_throttle)
        dsfc_dmach = (
            self.sfc_coeff[1]
            + 2 * self.sfc_coeff[4] * x_shared[2]
            + 2 * x_shared[1] * self.sfc_coeff[5]
            + 2 * throttle * self.sfc_coeff[6]
        )
        return dsfc_dmach

    def __dadimthrottle_dthrottle(self, x_3):
        """Compute partial derivative of adim throttle of polynomial function wrt
        throttle.

        :param x_3: local design variables
        :type x_3: numpy array
        :returns:  dadimthrottle_dthrottle
        :rtype: numpy array
        """
        return self.base.derive_normalize_s(self.throttle_initial, x_3[0])

    def __compute_dadimh_dh(self, x_shared):
        """Compute partial derivative of adim throttle of polynomial function wrt
        altitude.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns:  dadimh_dh
        :rtype: numpy array
        """
        return self.base.derive_normalize_s(self.h_initial, x_shared[1])

    def __compute_dadimmach_dmach(self, x_shared):
        """Compute partial derivative of adim throttle of polynomial function wrt Mach
        number.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns:  dadimthrottle_dthrottle
        :rtype: numpy array
        """
        return self.base.derive_normalize_s(self.mach_initial, x_shared[2])

    def __initialize_jacobian(self, true_cstr):
        """Initialization of jacobian matrix.

        :param true_cstr:
        :type true_cstr: logical
        :returns:  jacobian
        :rtype: dict of dict of numpy array
        """
        # Jacobian matrix as a dictionary
        jacobian = {"y_3": {}, "g_3": {}, "y_31": {}, "y_32": {}, "y_34": {}}

        jacobian["y_3"]["x_3"] = zeros((3, 1), dtype=self.dtype)
        jacobian["y_3"]["x_shared"] = zeros((3, 6), dtype=self.dtype)
        jacobian["y_3"]["y_23"] = zeros((3, 1), dtype=self.dtype)
        if not true_cstr:
            n_constraints = 4
        else:
            n_constraints = 3
        jacobian["g_3"]["x_3"] = zeros((n_constraints, 1), dtype=self.dtype)
        jacobian["g_3"]["x_shared"] = zeros((n_constraints, 6), dtype=self.dtype)
        jacobian["g_3"]["y_23"] = zeros((n_constraints, 1), dtype=self.dtype)

        return jacobian

    def derive_blackbox_propulsion(self, x_shared, y_23, x_3, true_cstr=False):
        """Compute jacobian matrix of propulsion analysis.

        :param x_shared: shared design variable vector:

            - x_shared[0]: thickness/chord ratio
            - x_shared[1]: altitude
            - x_shared[2]: Mach
            - x_shared[3]: aspect ratio
            - x_shared[4]: wing sweep
            - x_shared[5]: wing surface area

        :type x_shared: numpy array
        :param y_23: shared variables coming from blackbox_aerodynamics (drag)
        :type y_23: numpy array
        :param x_3: power/propulsion design variable (throttle setting)
        :type x_3: numpy array
        :param true_cstr: Default value = False)
        :returns: jacobian : Jacobian matrix
        :rtype: dict(dict(ndarray))
        """
        # Jacobian matrix as a dictionary
        jacobian = self.__initialize_jacobian(true_cstr)

        dg_3_dx_3 = zeros((3, 1), dtype=self.dtype)
        dg_3_dxs = zeros((3, 6), dtype=self.dtype)
        dg_3_dy_23 = zeros((3, 1), dtype=self.dtype)

        drag = y_23[0]

        esf = self.compute_esf(y_23[0], x_3[0])

        # dSFC_dthrottle
        jacobian["y_3"]["x_3"][0, 0] = self.compute_dsfc_dthrottle(x_shared, x_3[0])

        # dESF_dthrottle
        jacobian["y_3"]["x_3"][2, 0] = self.compute_desf_dthrottle(drag, x_3[0])
        # dengineweight_dthrottle
        jacobian["y_3"]["x_3"][1, :] = self.compute_dengineweight_dvar(
            esf, jacobian["y_3"]["x_3"][2, 0]
        )
        # dSFC_d(t/c) = 0
        # dESF_d(t/c) = 0
        # dengineweight_d(t/c) = 0

        # dSFC_dh
        jacobian["y_3"]["x_shared"][0, 1] = self.compute_dsfc_dh(x_shared, x_3[0])
        # dESF_dh= 0.0
        # dengineweight_dh= 0.0

        # dSFC_dM
        jacobian["y_3"]["x_shared"][0, 2] = self.compute_dsfc_dmach(x_shared, x_3[0])
        # dESF_dM= 0.0
        # dengineweight_dM= 0.0

        #        jacobian['y_3']['x_shared'][:, 2:] = 0.0

        # dSFC_ddrag
        #        jacobian['y_3']['y_23'][0, 0] = 0.0
        # dESF_ddrag
        jacobian["y_3"]["y_23"][2, 0] = self.compute_desf_ddrag(x_3[0])
        # dengineweight_ddrag
        jacobian["y_3"]["y_23"][1, :] = self.compute_dengineweight_dvar(
            esf, jacobian["y_3"]["y_23"][2, 0]
        )

        # dtemp_ddrag
        #        jacobian['g_3']['y_23'][0, 0] = 0.0
        s_initial, s_new, flag, bound = self.__set_coeff_temp(x_shared, x_3)
        _, ai_coeff, aij_coeff, s_shifted = self.base.derive_poly_approx(
            s_initial, s_new, flag, bound
        )

        #        dadimtemp_dtemp = self.__compute_dadimtemp_dtemp(x_shared)

        # g_3[0, :] = ESF
        dg_3_dx_3[0, :] = jacobian["y_3"]["x_3"][2, :]
        # dtemp_dthrottle
        dg_3_dx_3[1, 0] = self.__dadimthrottle_dthrottle(x_3) * (
            ai_coeff[2]
            + aij_coeff[2, 0] * s_shifted[0]
            + aij_coeff[2, 1] * s_shifted[1]
            + aij_coeff[2, 2] * s_shifted[2]
        )
        # d(throttle-throttle_ua)_dthrottle
        dg_3_dx_3[2, 0] = self.compute_dthrconst_dthrottle(x_shared)

        # g_3[0, :] = ESF
        dg_3_dxs[0, :] = jacobian["y_3"]["x_shared"][2, :]

        # dtemp_d(t/c)= 0.0
        # d(throttle-throttle_ua)_d(t/c)= 0.0

        # dtemp_dh
        dg_3_dxs[1, 1] = self.__compute_dadimh_dh(x_shared) * (
            ai_coeff[1]
            + aij_coeff[1, 0] * s_shifted[0]
            + aij_coeff[1, 1] * s_shifted[1]
            + aij_coeff[1, 2] * s_shifted[2]
        )
        # d(throttle-throttle_ua)_dh
        dg_3_dxs[2, 1] = self.compute_dthrcons_dh(x_shared, x_3[0])

        # dtemp_dM
        dg_3_dxs[1, 2] = self.__compute_dadimmach_dmach(x_shared) * (
            ai_coeff[0]
            + aij_coeff[0, 0] * s_shifted[0]
            + aij_coeff[0, 1] * s_shifted[1]
            + aij_coeff[0, 2] * s_shifted[2]
        )
        # d(throttle-throttle_ua)_dM
        dg_3_dxs[2, 2] = self.compute_dthrconst_dmach(x_shared, x_3[0])

        jacobian = self.__set_coupling_jacobian(jacobian)

        # THIS SECTION COMPUTES SFC, ESF, AND ENGINE WEIGHT
        dg_3_dxs[0, :] = jacobian["y_3"]["x_shared"][2, :]
        dg_3_dy_23[0, :] = jacobian["y_3"]["y_23"][2, :]

        if true_cstr:
            jacobian["g_3"]["x_3"] = dg_3_dx_3
            jacobian["g_3"]["x_shared"] = dg_3_dxs
            jacobian["g_3"]["y_23"] = dg_3_dy_23
        else:
            for i_g_jac, i_orig in enumerate([0, 0, 2, 1]):
                jacobian["g_3"]["x_shared"][i_g_jac, :] = dg_3_dxs[i_orig, :]
                jacobian["g_3"]["y_23"][i_g_jac, :] = dg_3_dy_23[i_orig, :]
                jacobian["g_3"]["x_3"][i_g_jac, :] = dg_3_dx_3[i_orig, :]
                if i_g_jac == 1:
                    jacobian["g_3"]["x_shared"][i_g_jac, :] *= -1
                    jacobian["g_3"]["y_23"][i_g_jac, :] *= -1
                    jacobian["g_3"]["x_3"][i_g_jac, :] *= -1
        return jacobian

    @staticmethod
    def __set_coupling_jacobian(jacobian):
        """Set jacobian of coupling variables."""
        jacobian["y_31"]["x_3"] = atleast_2d(jacobian["y_3"]["x_3"][1, :])
        jacobian["y_31"]["x_shared"] = atleast_2d(jacobian["y_3"]["x_shared"][1, :])
        jacobian["y_31"]["y_23"] = atleast_2d(jacobian["y_3"]["y_23"][1, :])

        jacobian["y_32"]["x_3"] = atleast_2d(jacobian["y_3"]["x_3"][2, :])
        jacobian["y_32"]["x_shared"] = atleast_2d(jacobian["y_3"]["x_shared"][2, :])
        jacobian["y_32"]["y_23"] = atleast_2d(jacobian["y_3"]["y_23"][2, :])

        jacobian["y_34"]["x_3"] = atleast_2d(jacobian["y_3"]["x_3"][0, :])
        jacobian["y_34"]["x_shared"] = atleast_2d(jacobian["y_3"]["x_shared"][0, :])
        jacobian["y_34"]["y_23"] = atleast_2d(jacobian["y_3"]["y_23"][0, :])
        return jacobian



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The API for uncertainty quantification and management."""
from __future__ import division, unicode_literals

from typing import Iterable, List, Optional, Sequence

from gemseo.algos.parameter_space import ParameterSpace
from gemseo.core.dataset import Dataset
from gemseo.core.discipline import MDODiscipline
from gemseo.uncertainty.distributions.distribution import Distribution
from gemseo.uncertainty.sensitivity.analysis import SensitivityAnalysis  # noqa: F401
from gemseo.uncertainty.statistics.statistics import Statistics


def get_available_distributions():  # type: (...) -> List[str]
    """Get the available distributions."""
    from gemseo.uncertainty.distributions.factory import DistributionFactory

    factory = DistributionFactory()
    return factory.available_distributions


def create_distribution(
    variable,  # type: str
    distribution_name,  # type: str,
    dimension=1,  # type: int
    **options
):  # type: (...) -> Distribution
    """Create a distribution.

    Args:
        variable: The name of the random variable.
        distribution_name: The name of a class
            implementing a probability distribution,
            e.g. 'OTUniformDistribution' or 'SPDistribution'.
        dimension: The dimension of the random variable.
        **options: The distribution options.

    Examples:
        >>> from gemseo.uncertainty.api import create_distribution
        >>>
        >>> distribution = create_distribution(
        ...     "x", "OTNormalDistribution", dimension=2, mu=1, sigma=2
        ... )
        >>> print(distribution)
        Normal(mu=1, sigma=2)
        >>> print(distribution.mean, distribution.standard_deviation)
        [1. 1.] [2. 2.]
        >>> samples = distribution.get_sample(10)
        >>> print(samples.shape)
        (10, 2)
    """
    from gemseo.uncertainty.distributions.factory import DistributionFactory

    factory = DistributionFactory()
    return factory.create(
        distribution_name, variable=variable, dimension=dimension, **options
    )


def get_available_sensitivity_analyses():  # type: (...) -> List[str]
    """Get the available sensitivity analyses."""
    from gemseo.uncertainty.sensitivity.factory import SensitivityAnalysisFactory

    factory = SensitivityAnalysisFactory()
    return factory.available_sensitivity_analyses


def create_statistics(
    dataset,  # type: Dataset
    variables_names=None,  # type: Optional[Iterable[str]]
    tested_distributions=None,  # type: Optional[Sequence[str]]
    fitting_criterion="BIC",  # type: str
    selection_criterion="best",
    level=0.05,  # type: float
    name=None,  # type: Optional[str]
):  # type: (...) -> Statistics
    """Create a statistics toolbox, either parametric or empirical.

    If parametric, the toolbox selects a distribution from candidates,
    based on a fitting criterion and on a selection strategy.

    Args:
        dataset: A dataset.
        variables_names: The variables of interest.
            If None, consider all the variables from dataset.
        tested_distributions: The names of
            the tested distributions.
        fitting_criterion: The name of a goodness-of-fit criterion,
            measuring how the distribution fits the data.
            Use :meth:`.ParametricStatistics.get_criteria`
            to get the available criteria.
        selection_criterion: The name of a selection criterion
            to select a distribution from candidates.
            Either 'first' or 'best'.
        level: A test level,
            i.e. the risk of committing a Type 1 error,
            that is an incorrect rejection of a true null hypothesis,
            for criteria based on a test hypothesis.
        name: A name for the statistics toolbox instance.
            If None, use the concatenation of class and dataset names.

    Returns:
        A statistics toolbox.

    Examples:
        >>> from gemseo.api import (
        ...     create_discipline,
        ...     create_parameter_space,
        ...     create_scenario
        ... )
        >>> from gemseo.uncertainty.api import create_statistics
        >>>
        >>> expressions = {"y1": "x1+2*x2", "y2": "x1-3*x2"}
        >>> discipline = create_discipline(
        ...     "AnalyticDiscipline", expressions_dict=expressions
        ... )
        >>> discipline.set_cache_policy(discipline.MEMORY_FULL_CACHE)
        >>>
        >>> parameter_space = create_parameter_space()
        >>> parameter_space.add_random_variable(
        ...     "x1", "OTUniformDistribution", minimum=-1, maximum=1
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x2", "OTNormalDistribution", mu=0.5, sigma=2
        ... )
        >>>
        >>> scenario = create_scenario(
        ...     [discipline],
        ...     "DisciplinaryOpt",
        ...     "y1",
        ...     parameter_space,
        ...     scenario_type="DOE"
        ... )
        >>> scenario.execute({'algo': 'OT_MONTE_CARLO', 'n_samples': 100})
        >>>
        >>> dataset = discipline.cache.export_to_dataset()
        >>>
        >>> statistics = create_statistics(dataset)
        >>> mean = statistics.mean()
    """
    from gemseo.uncertainty.statistics.empirical import EmpiricalStatistics as EmpStats
    from gemseo.uncertainty.statistics.parametric import (
        ParametricStatistics as ParamStats,
    )

    if tested_distributions is None:
        statistical_analysis = EmpStats(dataset, variables_names, name)
    else:
        statistical_analysis = ParamStats(
            dataset,
            tested_distributions,
            variables_names,
            fitting_criterion,
            level,
            selection_criterion,
            name,
        )
    return statistical_analysis


def create_sensitivity_analysis(
    analysis,  # type: str
    discipline,  # type: MDODiscipline
    parameter_space,  # type: ParameterSpace
    **options
):  # type: (...) -> SensitivityAnalysis
    """Create the sensitivity analysis.

    Args:
        analysis: The name of a sensitivity analysis class.
        discipline: A discipline.
        parameter_space: A parameter space.
        **options: The DOE algorithm options.

    Returns:
        The toolbox for these sensitivity indices.

    Examples:
        >>> from gemseo.api import create_discipline, create_parameter_space
        >>> from gemseo.uncertainty.api import create_sensitivity_analysis
        >>>
        >>> expressions = {"y1": "x1+2*x2", "y2": "x1-3*x2"}
        >>> discipline = create_discipline(
        ...     "AnalyticDiscipline", expressions_dict=expressions
        ... )
        >>>
        >>> parameter_space = create_parameter_space()
        >>> parameter_space.add_random_variable(
        ...     "x1", "OTUniformDistribution", minimum=-1, maximum=1
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x2", "OTNormalDistribution", mu=0.5, sigma=2
        ... )
        >>>
        >>> analysis = create_sensitivity_analysis(
        ...     "CorrelationIndices", discipline, parameter_space, n_samples=1000
        ... )
        >>> indices = analysis.compute_indices()
    """
    from gemseo.uncertainty.sensitivity.factory import SensitivityAnalysisFactory

    factory = SensitivityAnalysisFactory()

    name = analysis
    if "Analysis" not in name:
        name += "Analysis"
    name = name[0].upper() + name[1:]

    return factory.create(name, discipline, parameter_space, **options)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The factory to create the machine learning algorithms.

This module contains a factory to instantiate a :class:`.MLAlgo` from its class name.
This factory also provides a list of available machine learning algorithms and allows to
test if a machine learning algorithm is available.
"""
from __future__ import division, unicode_literals

import logging
import pickle
from os.path import join
from typing import List, Optional, Union

from gemseo.core.dataset import Dataset
from gemseo.core.factory import Factory
from gemseo.mlearning.core.ml_algo import MLAlgo, MLAlgoParameterType, TransformerType

LOGGER = logging.getLogger(__name__)


class MLAlgoFactory(object):
    """This factory instantiates a :class:`.MLAlgo` from its class name.

    The class can be either internal to |g| or external. In this second case, it can be
    either implemented in a module referenced in the "GEMSEO_PATH" or in a module The
    class can be either internal to |g| or external. In the second case, it can be
    either implemented in a module referenced in the GEMSEO_PATH environment variable or
    in a module starting with "gemseo_" and referenced in the PYTHONPATH environment
    variable.
    """

    def __init__(self):  # type: (...) -> None
        self.factory = Factory(MLAlgo, ("gemseo.mlearning",))

    def create(
        self,
        ml_algo,  # type: str
        **options  # type: Optional[Union[Dataset,TransformerType,MLAlgoParameterType]]
    ):  # type: (...) -> MLAlgo
        """Create an instance of a machine learning algorithm.

        Args:
            ml_algo: The name of a machine learning algorithm
                (its class name).
            **options: The options of the machine learning algorithm.

        Returns:
            The instance of the machine learning algorithm.
        """
        return self.factory.create(ml_algo, **options)

    @property
    def models(self):  # type: (...) -> List[str]
        """The available machine learning algorithms."""
        return self.factory.classes

    def is_available(
        self,
        ml_algo,  # type: str
    ):  # type: (...) -> bool
        """Check the availability of a machine learning algorithm.

        Args:
            ml_algo: The name of a machine learning algorithm (its class name).

        Returns:
            The availability of the machine learning algorithm.
        """
        return self.factory.is_available(ml_algo)

    def load(
        self,
        directory,  # type:str
    ):  # type: (...) -> MLAlgo
        """Load an instance of machine learning algorithm from the disk.

        Args:
            directory: The name of the directory
                containing an instance of a machine learning algorithm.

        Returns:
            The instance of the machine learning algorithm.
        """
        with open(join(str(directory), MLAlgo.FILENAME), "rb") as handle:
            objects = pickle.load(handle)
        model = self.factory.create(
            objects.pop("algo_name"),
            data=objects.pop("data"),
            **objects.pop("parameters")
        )
        for key, value in objects.items():
            setattr(model, key, value)
        model.load_algo(directory)
        return model
